; $Id: dm_calc_projection.pro,v 1.61.2.5 2013/05/29 15:09:24 ymqiu Exp $
;#######################################################################
;
; NAME:
;  dm_calc_projection
;
; PURPOSE:
;  calculate projection
;  apply masking, rebinning, and detector efficiency adjustment to data
;
; CATEGORY:
;  dcs_mslice
;
; HISTORY:
;  02/2010: added NXSPE support - Andrei Savici (saviciat@ornl.gov)   
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  May, 2013
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

@dm_filetools ;for dm_startend()

function dm_rangestring,values,step=step,unit=unit,tolerance=tolerance
    if n_elements(values) eq 0 then return,''
    if n_elements(step) eq 0   then step=0.0 else step=abs(step) 
    if n_elements(unit) eq 0   then unit=''
    if n_elements(tolerance) eq 0 then tolerance = step/1e4
    tmp =  values[uniq(values,sort(values))]
    n_tmp = n_elements(tmp)
    if step eq 0 then return,dm_to_string(tmp[0])+unit+((n_tmp gt 1)?' to '+dm_to_string(tmp[n_tmp-1])+unit:'')
    outstr = dm_to_string(tmp[0])+unit
    vstart = tmp[0] & cnt = 0L
    for i=1L,n_tmp-1 do begin
        cnt = cnt+1
        if abs(vstart+cnt*step-tmp[i]) gt tolerance then begin
           if abs(tmp[i-1]-vstart) gt tolerance then outstr = outstr+' to '+dm_to_string(tmp[i-1])+unit
           outstr = outstr+', '+dm_to_string(tmp[i])+unit
           vstart = tmp[i] & cnt=0L
        endif   
    endfor
    if abs(tmp[i-1]-vstart) gt tolerance then outstr = outstr+' to '+dm_to_string(tmp[i-1])+unit
    return,outstr
end


;This function calculates the MACS monitor lambda/2 correction
;parameter:
;     Ei:       [ndat], incident energy
;     monitor:  [ndat], corresponding monitor counts
;     cfx:      [ndat], corresponding CFX status 0-no filter, correction will be applied
;keywords:
;     extra keywords for poly_fit
;return:
;     [ndat], new monitor counts
function dm_macs_lambda2correction,Ei,monitor,cfx,_ref_extra=_extra
    index  = where(cfx eq 0,count)
    if count eq 0 then return,monitor
    ;factor is from Jose's Intensitiesl_l2_Sim_nov.txt file
    factor = [[2,   2.378025665d],[2.5, 1.793602670d],[3,   1.490934555d],[3.5, 1.252089079d],[4,   1.127126095d],[4.5, 0.959301163d],[5,   0.792585170d],[5.5, 0.649712929d],$
              [6,   0.582762512d],[6.5, 0.521160485d],[7,   0.483133752d],[7.5, 0.444736852d],[8,   0.416914118d],[8.5, 0.390288542d],[9,   0.362685682d],[9.5, 0.347918053d],$
              [10,  0.339471838d],[10.5,0.327112776d],[11,  0.325793541d],[11.5,0.305038794d],[12,  0.292795409d],[12.5,0.284034166d],[13,  0.264029518d],[13.5,0.268863169d],$
              [14,  0.259363234d],[14.5,0.247959631d],[15,  0.246067755d],[15.5,0.237768686d],[16,  0.225825977d],[16.5,0.228803554d],[17,  0.229217028d],[17.5,0.235338426d],$
              [18,  0.236590056d],[18.5,0.233790439d],[19,  0.242718897d],[19.5,0.242768485d],[20,  0.242982347d]]
    parm   = poly_fit(factor[0,*],factor[1,*],7,_extra=_extra)
    newmon = monitor
    newmon[index] = newmon[index]/(1.0+0.5*float(poly(Ei[index],parm)))
    return,newmon
end

;keywords:
;     temperature: if present, no prompt for temperautre, useful in scripts
;     dwu2:        if present, no prompt for Debye-Waller factro correction u^2, useful in scripts
pro dcs_mslice::dm_calc_projection,temperature=temperature,dwu2=dwu2,debug=debug
    self.error = 1  ;initialize the error flag, it will be cleared at the end of the program
    if keyword_set(debug) then begin
       print,'**Calculate projection...'
       current = systime(/sec)
    end

    case self.instrname of
      'dcs':begin     ;dcs
            if (~ ptr_valid(self.detPosPtr)) then self.detPosPtr=dm_load_detpos(/dcs,adjust2th=self.tthcorr)
            end
      'macs':begin    ;macs
            if (~ ptr_valid(self.detPosPtr)) then self.detPosPtr=dm_load_detpos(/macs)
            end
      'spe':begin     ;spe
            if (~ ptr_valid(self.detPosPtr)) then begin
               self->dm_load_phx,phxfile=phxfile
               if strlen(phxfile) ne 0 then self.phxfile=phxfile
            endif
            end
      'nxspe':begin   ;nxspe
            if (~ ptr_valid(self.detPosPtr)) then begin
               self->dm_load_phx,phxfile=phxfile
            endif
            end
      'inx':begin     ;inx
            if (~ ptr_valid(self.detPosPtr)) then begin
               self.detPosPtr = dm_load_detpos(/neat)
               ok = dialog_message('Detector information is not available, using NEAT instrument info instead.',title='Please note:',dialog_parent=self.tlb,/center)
            endif
            end
      'wand':begin    ;wand
            if (~ ptr_valid(self.detPosPtr)) then self.detPosPtr=dm_load_detpos(/wand,adjust2th=self.tthcorr)
            end
      else:
    endcase

    if (~ ptr_valid(self.detPosPtr)) then begin
       info = ['Detector information is not available. The projection calculation is aborted.']
       if ((self.instrname eq 'nxspe') or (self.instrname eq 'spe')) and (~ptr_valid(self.dataStrPtr)) then $
       info = [info,'','You can load an existing '+self.instrname+' file and clear data to load the detector information.']
       ok   = dialog_message(info,/error,dialog_parent=self.tlb,/center)
       return
    endif

    extravaxis = (self.extravaxis_yn and (self.instrname ne 'macs') and (self.instrname ne 'wand'))

    ;check if the common input parameters are all set for single crystal and diffuse scattering type
    if self.samp_typ ne 0 then begin
       if total(finite(self.latt_parm,/nan)) ne 0 then begin
          mesg = 'Missing lattice parameters.'
       endif
       if total([finite(self.orie_u,/nan),finite(self.orie_v,/nan),finite(self.orie_psi,/nan)]) ne 0 then begin
          if n_elements(mesg) eq 0 then $
             mesg = 'Missing crystal orientation parameters.'  $
          else $
             mesg = [mesg,'Missing crystal orientation parameters.']
       endif
       if (total([finite(self.view_u1,/nan),finite(self.view_u2,/nan)]) ne 0) or ((total(finite(self.view_u3,/nan)) ne 0) and extravaxis) then begin
          if n_elements(mesg) eq 0 then $
             mesg = 'Missing viewing axes parameters.'  $
          else $
             mesg = [mesg,'Missing viewing axes parameters.']
       endif
       if n_elements(mesg) gt 0 then begin
          ok = dialog_message(mesg,/error,dialog_parent=self.tlb,/center)
          return
       endif
       if (self.instrname eq 'macs') or (self.instrname eq 'wand') and ~self->viewonscattplane() then begin
          ok = dialog_message('The viewing axes are not on the scattering plane.',dialog_parent=self.tlb,/center) ;not fatal, allow it to continue
       endif
    endif

    if (strlen(self.userf_str) ne 0) then begin
       userf_str = self.userf_str 
       ;remove comment lines, just in case
       repeat begin
          tmp1 = stregex(userf_str,'^(;[^&]*&+)',/subexpr,length=len)                                     ;the comment sign either starts at the very beginning
          if len[1] le 0 then tmp1 = stregex(userf_str,'&(;[^&]*&+)',/subexpr,length=len)                 ;or has to immediately follow a & sign
          if len[1] le 0 then tmp1 = stregex(userf_str,'&[^&"'+"'"+']*(;[^&]*&*)&',/subexpr,length=len)   ;or no ' or " before comment sign
          if len[1] le 0 then tmp1 = stregex(userf_str,'^[^&"'+"'"+']*(;[^&]*&*)&',/subexpr,length=len)  
          if len[1] le 0 then tmp1 = stregex(userf_str,'(;[^&]*$)',/subexpr,length=len)                   ;or ends with a comment 
          if len[1] gt 0 then userf_str = strmid(userf_str,0,tmp1[1])+strmid(userf_str,tmp1[1]+len[1],strlen(userf_str)-tmp1[1]-len[1]) ;remove it
       endrep until len[1] le 0 
       userf_str = strtrim(userf_str,2)
    endif else userf_str = ''

    if (~ ptr_valid(self.dataStrPtr)) then begin    ;no data loaded yet
       if self.instrname eq 'macs' then begin
          eief = 5.0
          if (finite(self.eief)) then $
             if self.eief gt 0 then eief=dm_to_number(dm_to_string(self.eief,resolution=3))
          fts = ['from','to','step']
          name = ['fixed '+(['Ei','Ef'])[self.instrgeom]+' (meV):','E '+fts+' (meV):','Kidney '+fts+':']
          default = [eief,self.estartend,self.kidney]    
          info = ['Leave kidney range blank for full allowable kidney range.','Kidney offset = '+dm_to_string(self.kidney_offset,resolution=3)+string('b0'XB)]
          if self.samp_typ eq 1 then begin
             name = [name,'A3 '+fts+':']
             default = [default,self.psi[0:2]]
             if self.datr_yn then begin
                name    = [name,'Dark Angle:'] 
                default = [default,self.datr_psi]
                info    = [info,'Dark angle position is relative to the crystal axis u.','Counterclockwise rotation is positive.']
             endif
          endif
          notok = 1b
          while(notok) do begin
               ans = dm_dialog_input(name,/float,default=default,dialog_parent=self.tlb,/align_center,info=info,cancel=cancel)
               if keyword_set(cancel) then return
               if total(finite(ans,/nan)) eq total(finite(ans[4:5],/nan)) then begin
                  if ans[0] gt 0 then notok = 0b
                  if (total(finite(ans[4:5],/nan)) ne 0) and (ans[6] eq 0) then notok = 1b
               endif
          endwhile
          kid_fnt = finite(ans[4:5])
          ans[[3,6]] = abs(ans[[3,6]])
          if ans[2] lt ans[1] then ans[1:2] = ans[[2,1]]
          if ans[3] eq 0 then ans[2] = ans[1] $
          else if ans[1] eq ans[2] then ans[3] = 0
          if kid_fnt[0] and kid_fnt[1] then begin
             if ans[5] lt ans[4] then ans[4:5] = ans[[5,4]]
             if ans[6] eq 0 then ans[5] = ans[4] $
             else if ans[4] eq ans[5] then ans[6] = 0
          endif else $  ;get the kidney limit
             tmp = dm_check_macskidney(0,5,limits=limits)
          eief           = ans[0]
          self.eief      = ans[0] 
          self.estartend = ans[1:3]
          self.kidney    = ans[4:6]
          nen            = 1>((self.estartend[2] eq 0)?1:round(1+(self.estartend[1]-self.estartend[0])/self.estartend[2]))
          if self.samp_typ eq 1 then begin
             ans[9] = abs(ans[9])
             if ans[8] lt ans[7] then ans[7:8] = ans[[8,7]]
             if ans[9] eq 0 then ans[8] = ans[7] $
             else if ans[7] eq ans[8] then ans[9] = 0
             self.psi[0:2] = ans[7:9]
             na3 = 1>((self.psi[2] eq 0)?1:round(1+(self.psi[1]-self.psi[0])/self.psi[2]))  
             if self.datr_yn then self.datr_psi = ans[10]
          endif else na3 = 1
          for i=0L,na3-1L do begin
              tmp_a3 = self.psi[0]+self.psi[2]*i
              for j=0L,nen-1L do begin
                  tmp_en = self.estartend[0]+self.estartend[2]*j
                  if self.instrgeom eq 0 then tmp_ei = eief $ ;direct geometry
                  else tmp_ei = eief+tmp_en                   ;inverse geometry
                  kidran = self.kidney[0:1] 
                  if ~kid_fnt[0] then kidran[0] = interpol(limits[*,1],limits[*,0],tmp_ei)-self.kidney_offset+0.001
                  if ~kid_fnt[1] then kidran[1] = interpol(limits[*,2],limits[*,0],tmp_ei)-self.kidney_offset-0.001
                  nkid = 1>((self.kidney[2] eq 0)?1:round(1+(kidran[1]-kidran[0])/self.kidney[2]))
                  for k=0L,nkid-1L do begin
                      tmp_kid = (kidran[0])>(kidran[0]+self.kidney[2]*k)<(kidran[1])
                      if n_elements(a3) eq 0 then begin
                         a3     = tmp_a3
                         en     = tmp_en
                         Ei     = tmp_ei
                         kidney = tmp_kid
                      endif else begin
                         a3     = [a3,tmp_a3]
                         en     = [en,tmp_en]
                         Ei     = [Ei,tmp_ei]
                         kidney = [kidney,tmp_kid]
                      endelse
                  endfor
              endfor
          endfor
          nen = n_elements(en)
          Ef  = Ei-en
          psi = 90.0-temporary(a3)
          ;check kidney
          index = dm_check_macskidney(kidney,Ei,count,complement=bindex,offset=self.kidney_offset)
          if (count eq 0) then begin
             ok = dialog_message('All kidney angles are out of valid range.',/error,dialog_parent=self.tlb,/center)
             return 
          endif else if count ne nen then begin
             bad_en  = en[bindex]
             bad_kid = kidney[bindex]
             Ei  = Ei[index]
             Ef  = Ef[index]
             en  = en[index]
             psi = psi[index]
             kidney = kidney[index]
             tmp = bad_en[uniq(bad_en,sort(bad_en))]
             if n_elements(mesg) eq 0 then mesg = 'The following kidney angles are out of valid range.' $
             else mesg = [mesg,'','The following kidney angles are out of valid range.']
             for i=0L,n_elements(tmp)-1L do begin
                 index = where(bad_en eq tmp[i])
                 tmpmin = min(bad_kid[index],max=tmpmax)
                 mesg = [mesg,'E='+dm_to_string(tmp[i])+' meV, kidney='+dm_rangestring(bad_kid[index],step=self.kidney[2],unit=string('b0'XB),tolerance=0.0001)]
             endfor
             mesg = [mesg,'These kidney angles have been removed from projection calculation.','']
             nen = count
          endif
          index = 0  & bindex = 0
       endif else begin
          ;ask for Ei or Ef according to instrument geometry
          eief = 5.0
          if (finite(self.eief)) then $
             if self.eief gt 0 then eief=self.eief
          if self.samp_typ eq 2 then begin             ;diffuse scattering single E
             case self.e_bin[self.samp_typ] of
               0:  en = 0.0                                      ;use elastic energy for all time channels
               1:  en = 0.0                                      ;use elastic energy for elastic peak
               2:  en = (self.e_range[0]+self.e_range[1])/2.0    ;use the mean energy for specified energy range
               else: en = 0.0
             endcase
             askeran = 0b
             if self.instrname eq 'wand' then $
                info = ['detector offset angle='+dm_to_string(-(*self.detPosPtr).two_theta[0],resolution=3)+string('b0'XB)] $
             else $
                info = ['E='+dm_to_string(en)+' meV']
          endif else if self.samp_typ eq 3 then begin  ;diffuse scattering multiple E
             askeran = 1b
             case self.e_bin[3] of
               0:  info = (['-Ei< E <Ei','-Ef< E <Ef'])[self.instrgeom]
               1:  begin
                   info = ['E~0 meV']
                   askeran = 0b
                   end
               2:  begin
                   info = [dm_to_string(self.e_range[0])+'< E <'+dm_to_string(self.e_range[1])]
                   if strlen(userf_str) ne 0 then self.estartend[0:1] = self.e_range[0:1]
                   end
               else:
             endcase
          endif else begin
             info = (['-Ei< E <Ei','-Ef< E <Ef'])[self.instrgeom]
             askeran = 1b
          endelse
          if (self.instrname eq 'dcs') or (self.instrname eq 'wand') then begin    ;for DCS convert energy(meV) to wavelength(angstrom)
             eief  = sqrt(81.8042/eief)
             title = "Incident wavelength:"
          endif else $
             title = (["Incident energy:","Final energy:"])[self.instrgeom]
          notok = 1b
          while(notok) do begin
             if askeran and strlen(userf_str) ne 0 then begin
                if (self.instrname eq 'dcs') or (self.instrname eq 'wand') then begin
                   eief = dm_dialog_input(['wavelength ('+string('c5'XB)+'):','E_start:','E_end:','n_E:'],/float,default=[finite(eief[0])?round(eief[0]*1000.0,/L64)/1000.0:eief[0],self.estartend],$
                          dialog_parent=self.tlb,info='E_start< E <E_end, n_E is the sampling number.',/lambda2e,title=title,/align_center)
                endif else begin
                   eief = dm_dialog_input([(['Ei','Ef'])[self.instrgeom]+'(meV):','E_start:','E_end:','n_E:'],/float,default=[finite(eief[0])?round(eief[0]*1000.0,/L64)/1000.0:eief[0],self.estartend],$
                          dialog_parent=self.tlb,info='E_start< E <E_end, n_E is the sampling number.',/e2lambda,title=title,/align_center)
                endelse
             endif else begin
                if (self.instrname eq 'dcs') or (self.instrname eq 'wand') then begin
                   eief = dm_dialog_input('wavelength ('+string('c5'XB)+'):',/float,default=finite(eief)?round(eief*1000.0,/L64)/1000.0:eief,dialog_parent=self.tlb,info=info,/lambda2e,title=title,/align_center)
                endif else begin
                   eief = dm_dialog_input((['Ei','Ef'])[self.instrgeom]+'(meV):',/float,title=title,default=finite(eief)?round(eief*1000.0,/L64)/1000.0:eief,dialog_parent=self.tlb,info=info,/e2lambda,/align_center)
                endelse
             endelse
             if (total(finite(eief)) eq n_elements(eief)) then $
                if eief[0] gt 0 then notok = 0b
          endwhile
          if (self.instrname eq 'dcs') or (self.instrname eq 'wand') then $    ;for DCS and WAND convert wavelength to energy
             eief[0] =  81.8042/(eief[0]^2)
          self.eief = eief[0]  ;save ei or ef information
          if askeran and strlen(userf_str) ne 0 then begin
             if eief[3] lt 1 then eief[3] = 81
             if self.instrgeom eq 0 then eief[2] = min([eief[2],eief[0]]) $
             else eief[1] = max([eief[1],-eief[0]])
             en = eief[1] + findgen(eief[3])/(eief[3]-1)*(eief[2]-eief[1]) +(2*self.instrgeom-1)*1e-6
             self.estartend = eief[1:3]
             eief = eief[0]
          endif else begin
             nnen = 81.           ;sampling number
             if self.samp_typ eq 2 then begin             ;diffuse scattering single E
                if (self.instrgeom eq 0) and (en gt eief) then begin
                   ok = dialog_message('Incident energy is less than the energy transfer: E>Ei. E=0 is used.',/error,dialog_parent=self.tlb,/center)
                   en = 0.0
                endif
                if (self.instrgeom eq 1) and (en lt -eief) then begin
                   ok = dialog_message('Neutron energy gain is more than the final energy: E<-Ef. E=0 is used.',/error,dialog_parent=self.tlb,/center)
                   en = 0.0
                endif
             endif else if self.samp_typ eq 3 then begin  ;diffuse scattering multiple E
                case self.e_bin[3] of
                  0:  en = (findgen(nnen)/((nnen-1)/2.)-1)*eief+(2*self.instrgeom-1)*1e-6         ;[-Ei,...,Ei] or [-Ef,...Ef]
                  1:  begin
                      nnen = 21.
                      ;use empirical DCS energy resolution as the energy range
                      dE = 0.02008*eief^(1.4735)
                      en = (findgen(nnen)/((nnen-1)/2.)-1)*dE
                      end
                  2:  begin
                      if self.instrgeom eq 0 then begin
                         if min(self.e_range) gt eief then begin
                            ok = dialog_message('Incident energy is smaller than the specified energy range. Elastic peak is used.',/error,dialog_parent=self.tlb,/center)
                            ;use empirical DCS energy resolution as the energy range
                            dE = 0.02008*eief^(1.4735)
                            en = (findgen(nnen)/((nnen-1)/2.)-1)*dE
                         endif else begin
                            eran = [self.e_range[0],min([self.e_range[1],eief])]
                            en   = eran[0] + findgen(nnen)/(nnen-1)*(eran[1]-eran[0])-1e-6
                         endelse
                      endif else begin
                         if max(self.e_range) lt -eief then begin
                            ok = dialog_message('Neutron energy gain in the specified energy range is more than the final energy. Elastic peak is used.',/error,dialog_parent=self.tlb,/center)
                            ;use empirical DCS energy resolution as the energy range
                            dE = 0.02008*eief^(1.4735)
                            en = (findgen(nnen)/((nnen-1)/2.)-1)*dE
                         endif else begin
                            eran = [max(self.e_range[0],-eief),self.e_range[1]]
                            en   = eran[0] + findgen(nnen)/(nnen-1)*(eran[1]-eran[0])+1e-6
                         endelse
                      endelse
                      end
                  else:
                endcase
             endif else begin
                en = (findgen(nnen)/((nnen-1)/2.)-1)*eief +(2*self.instrgeom-1)*1e-6         ;[-Ei,...,Ei] or [-Ef,...Ef]
             endelse
          endelse
          if self.samp_typ ge 2 then begin ;psi range for diffuse scattering type
             notok = 1b
             if total(finite(self.psi[0:2])) gt 0 then psi=self.psi[0:2] else psi = [0,50,2]
             name = ['Start:','End:','Step:']
             if self.datr_yn then begin
                psi    = [psi,self.datr_psi]
                name   = [name,'Dark Angle:'] 
                dainfo = ['Dark angle position is relative to the crystal axis u.','Counterclockwise rotation is positive.']
             endif
             while(notok) do begin
                psi = dm_dialog_input(name,title='Psi range:',default=psi,/float,dialog_parent=self.tlb,info=dainfo)
                if (total(finite(psi[0:2],/nan)) eq 0) then notok = 0b
             endwhile
             psi[2]  = abs(psi[2])
             min_a  = min(psi[0:1],max=max_a)
             if psi[2] eq 0 then max_a = min_a $
             else if min_a eq max_a then psi[2] = 0
             self.psi[0:2] = [min_a,max_a,psi[2]]
             if self.datr_yn then self.datr_psi = psi[3]
             if psi[2] eq 0 then psi = min_a $
             else psi = min_a+findgen(fix((max_a-min_a)/psi[2])+1)*psi[2]
             n_file = n_elements(psi)
          endif
       endelse
       qty  = 0.
       err  = 0.
       ewid = 0.
       time = 0.
       data_temperature = 0.0 ;data temperature, for MACS and WAND
    endif else begin
       data_temperature = (*self.dataStrPtr).temperature ;data temperture, for MACS and WAND      
       if self.instrname eq 'macs' then begin
          qty = (*self.dataStrPtr).qty
          err = (*self.dataStrPtr).err
          Ei  = (*self.dataStrPtr).Ei
          Ef  = (*self.dataStrPtr).Ef
          psi = 90.0-((*self.dataStrPtr).rotation+self.orie_psi)  ;actually A3,self.orie_psi is the A3 offset angle, 90-a3 due to u definition and rotation definition  
          en  = Ei-Ef
          eief = self.eief
          kidney = (*self.dataStrPtr).kidney
          if self.macs_lamb2 then weight = dm_macs_lambda2correction(Ei,(*self.dataStrPtr).weight,(*self.dataStrPtr).cfx) $
          else weight = (*self.dataStrPtr).weight
          monfac = (self.macs_mon[self.mon_sumfit])/weight
          for i=0L,n_elements(monfac)-1 do begin
              qty[*,*,i] = qty[*,*,i]*monfac[i]
              err[*,*,i] = err[*,*,i]*monfac[i]
          endfor
          if ptr_valid(self.bgdata) then begin
             bgqty = (*self.bgdata).qty
             bgerr = (*self.bgdata).err
             bgang = 90.0-((*self.bgdata).rotation+self.orie_psi)
             bgen  = (*self.bgdata).Ei-(*self.bgdata).Ef
             bgkid = (*self.bgdata).kidney
             nbgen = n_elements(bgen)
             if self.macs_lamb2 then weight = dm_macs_lambda2correction((*self.bgdata).Ei,(*self.bgdata).weight,(*self.bgdata).cfx) $
             else weight = (*self.bgdata).weight
             monfac = (self.macs_mon[self.mon_sumfit])/weight
             for i=0L,nbgen-1L do begin
                 bgqty[*,*,i] = bgqty[*,*,i]*monfac[i]
                 bgerr[*,*,i] = bgerr[*,*,i]*monfac[i]
             endfor
          endif
          if ptr_valid(self.bgdata2) then begin
             bgqty2 = (*self.bgdata2).qty
             bgerr2 = (*self.bgdata2).err
             bgang2 = 90.0-((*self.bgdata2).rotation+self.orie_psi)
             bgen2  = (*self.bgdata2).Ei-(*self.bgdata2).Ef
             bgkid2 = (*self.bgdata2).kidney
             nbgen2 = n_elements(bgen2)
             if self.macs_lamb2 then weight = dm_macs_lambda2correction((*self.bgdata2).Ei,(*self.bgdata2).weight,(*self.bgdata2).cfx) $
             else weight = (*self.bgdata2).weight
             monfac = (self.macs_mon[self.mon_sumfit])/weight
             for i=0L,nbgen2-1L do begin
                 bgqty2[*,*,i] = bgqty2[*,*,i]*monfac[i]
                 bgerr2[*,*,i] = bgerr2[*,*,i]*monfac[i]
             endfor
             ;apply the empty can factor
             if total(finite(self.empcanfactor)) eq 2 then begin
                bgqty  = bgqty*self.empcanfactor[0]
                bgerr  = bgerr*self.empcanfactor[0]
                bgqty2 = bgqty2*self.empcanfactor[1]
                bgerr2 = bgerr2*self.empcanfactor[1]
             endif
          endif
          monfac = 0
          ewid = 1
       endif else begin
          ewid = (*self.dataStrPtr).ewid 
          eief = (*self.dataStrPtr).eief & self.eief = eief
          qty  = (*self.dataStrPtr).qty
          err  = (*self.dataStrPtr).err
          while (~ finite(eief)) do begin
               eief = dm_dialog_input((['Ei','Ef'])[self.instrgeom]+'(meV):',/float,default=eief,dialog_parent=self.tlb,/e2lambda)
               (*self.dataStrPtr).eief = eief
               self.eief = eief
          endwhile
          en   = (*self.dataStrPtr).e 
          time = (*self.dataStrPtr).time
          if self.samp_typ ge 2 then begin                    ;diffuse scattering type
             psi = (*self.dataStrPtr).rotation                 ;angle 2-rotation
             self.orie_offset = self.orie_psi-self.psi[3]      ;save the offset information
             psi = psi+self.orie_offset                         ;adjust the offset
             n_file = n_elements(psi)
             self.psi[0:2] = dm_startend(psi,resolution=2)
          endif else if self.samp_typ eq 1 then begin
             if finite((*self.dataStrPtr).rotation) then $
                self.orie_offset = self.orie_psi-(*self.dataStrPtr).rotation
          endif
          if ptr_valid(self.bgdata) then begin
             bgqty = (*self.bgdata).qty
             bgerr = (*self.bgdata).err
             bgang = (*self.bgdata).rotation
             if self.samp_typ ge 2 then n_bgfile = n_elements(bgang)
          endif
          if ptr_valid(self.bgdata2) then begin
             bgqty2 = (*self.bgdata2).qty
             bgerr2 = (*self.bgdata2).err
             bgang2 = (*self.bgdata2).rotation
             if self.samp_typ ge 2 then n_bgfile2 = n_elements(bgang2)
             ;apply the empty can factor
             if total(finite(self.empcanfactor)) eq 2 then begin
                bgqty  = bgqty*self.empcanfactor[0]
                bgerr  = bgerr*self.empcanfactor[0]
                bgqty2 = bgqty2*self.empcanfactor[1]
                bgerr2 = bgerr2*self.empcanfactor[1]
             endif
          endif
       endelse
    endelse

    det_tt  = (*self.detPosPtr).two_theta
    det_psi = (*self.detPosPtr).psi
    nen     = n_elements(en)
    ndet    = n_elements(det_tt)
   
    widget_control,/hourglass
    
    ;user function executed before empty can subtraction
    if (strlen(userf_str) ne 0) and (self.userf_before eq 2) then begin
       if self.userfwarning then begin
          if lmgr(/runtime) then usertmp = 'User Macro' else usertmp = 'User Macro/Function/Procedure'
          if n_elements(mesg) eq 0 then mesg = usertmp+' is specified.' $
          else mesg = [usertmp+' is specified.','',mesg]
          self.userfwarning = 0b
       endif
       ok = execute(userf_str,1,1)
       if ~ok then begin
          ok = dialog_message(!error_state.msg,/error,dialog_parent=self.tlb,/center)
          return  ;error encountered
       endif
       if n_elements(err) ne n_elements(qty) then begin ;in case qty and err are not both defined in user functions
          if n_elements(err) eq 1 then begin
             err = replicate(err,size(qty,/dimension))
          endif else begin
             help,qty,err,output=tmp
             ok = dialog_message(['Incompatible array sizes of intensity and error bar.','',tmp],/error,dialog_parent=self.tlb,/center)
             return
          endelse
       endif
    endif
    
    
    if ptr_valid(self.dataStrPtr) then begin
       if (self.samp_typ eq 0) and self.ftypecheckmask then begin
          i = 0L
          while(total(finite(qty[i,*])) eq 0) do i=i+1    ;in case the first time channel is masked
          qtydet =  reform(qty[i,*])                      ;save the unadjust qty info for later mask determination
       endif
       ;subtract overall background rate
       if ptr_valid(self.bgratePtr) then begin
          bgrate = (*self.bgratePtr)
          nrate  = n_elements(bgrate)
          if nrate gt 1 then begin
             dbgrate = bgrate[*,1]
             bgrate  = bgrate[*,0]
          endif else $
             dbgrate = 1.0      ;arbitrary error bar
          case self.instrname of
            'dcs':begin   ;dcs
                  if nrate eq 1 then begin    ;overall rate
                     if ptr_valid(self.eff) then tmp_eff=(*self.eff) else tmp_eff=fltarr(ndet)+1.
                     tmp_value = bgrate/3600./(self.monrate)*temporary(tmp_eff)
                  endif else $
                     tmp_value = bgrate/3600./(self.monrate)
                  if ptr_valid(self.bgdata2) and total(finite(self.empcanfactor)) eq 2 then factor = self.empcanfactor else factor = [1.,1.]   
                  if self.samp_typ ge 2 then begin      ;diffuse scattering type
                     if self.samp_typ eq 2 then begin   ;single energy
                        if self.eint_yn then $
                           tmp_value = tmp_value*total(ewid) $
                        else $
                           tmp_value = tmp_value
                        for i=0L,n_file-1 do qty[*,i]=qty[*,i]-tmp_value
                        if ptr_valid(self.bgdata) then $
                           for i=0L,n_bgfile-1 do bgqty[*,i]=bgqty[*,i]-tmp_value*factor[0]
                        if ptr_valid(self.bgdata2) then $
                           for i=0L,n_bgfile2-1 do bgqty2[*,i]=bgqty2[*,i]-tmp_value*factor[1]
                     endif else begin                    ;multiple energy
                        for i=0L,n_file-1L do begin
                            for j=0L,nen-1L do $
                                qty[j,*,i] = qty[j,*,i]-tmp_value
                        endfor
                        if ptr_valid(self.bgdata) then begin
                           for i=0L,n_bgfile-1L do begin
                               for j=0L,nen-1L do $
                                   bgqty[j,*,i] = bgqty[j,*,i]-tmp_value*factor[0]
                           endfor
                        endif
                        if ptr_valid(self.bgdata2) then begin
                           for i=0L,n_bgfile2-1L do begin
                               for j=0L,nen-1L do $
                                   bgqty2[j,*,i] = bgqty2[j,*,i]-tmp_value*factor[1]
                           endfor
                        endif
                     endelse
                  endif else begin
                     for i=0L,nen-1L do $
                         qty[i,*] = qty[i,*]-tmp_value
                     if ptr_valid(self.bgdata) then begin
                        for i=0L,nen-1L do $
                            bgqty[i,*] = bgqty[i,*]-tmp_value*factor[0]
                     endif
                     if ptr_valid(self.bgdata2) then begin
                        for i=0L,nen-1L do $
                          bgqty2[i,*] = bgqty2[i,*]-tmp_value*factor[1]
                     endif
                  endelse
                  tmp_value = 1     ;free up the memory
                  end
            else:
          endcase
       endif
       ;subtract detector dependent background data
       if ptr_valid(self.bgdetsPtr) then begin
          tmp_value = (*self.bgdetsPtr)
          if self.samp_typ ge 2 then begin
             if self.eint_yn then $
                tmp_value = tmp_value*total(ewid)
             if self.samp_typ eq 2 then begin
                for i=0L,n_file-1L do $
                    qty[*,i] = qty[*,i]-tmp_value
             endif else begin
                tmp_value = (fltarr(nen)+1.0)#tmp_value
                for i=0L,n_file-1L do $
                    qty[*,*,i] = qty[*,*,i]-tmp_value
             endelse
          endif else begin
             qty = qty-(fltarr(nen)+1.0)#tmp_value
          endelse
          tmp_value = 1     ;free up the memory
       endif
       ;subtract time channel background data
       if ptr_valid(self.bgtchnPtr) then begin
          tmp_emin  = min((*self.bgtchnPtr).emean,max=tmp_emax)
          tmp_nen   = n_elements((*self.bgtchnPtr).emean)
          tchan_ind = where(en ge tmp_emin and en le tmp_emax,tchan_count)
          if tchan_count gt tmp_nen then tchan_ind = tchan_ind[0:(tmp_nen-1)]   ;errorproof
          if tchan_count ne 0 then begin
             if self.samp_typ eq 3 then begin
                for i=0L,n_file-1L do $
                    for j=0L,ndet-1L do $
                        qty[tchan_ind,j,i] = qty[tchan_ind,j,i]-(*self.bgtchnPtr).qty
             endif else begin
                for i=0L,ndet-1L do $
                    qty[tchan_ind,i] = qty[tchan_ind,i]-(*self.bgtchnPtr).qty
             endelse
          endif
       endif else tchan_count = 0
       ;subtract empty can file
       if ptr_valid(self.bgdata) then begin
          if finite(self.ssfac) then ssfac=self.ssfac $
          else ssfac = 1.0     ;self-shielding factor
          if self.instrname eq 'macs' then begin
             for i=0L,nen-1 do begin
                 if self.macs_bgefirst then begin ;search order: En->kidney->A3
                    tmp1 = min(abs(bgen-en[i]),imin) & tmp1 = bgen[imin]
                    if self.empt_tol_optn eq 0 then index1 = where((bgen eq tmp1) or (abs(bgen-en[i]) le self.macs_tolerance[0])) $
                    else index1 = where(abs(bgen-tmp1) le self.macs_tolerance[0])
                    tmp1 = min(abs(bgkid[index1]-kidney[i]),imin) & tmp1 = bgkid[index1[imin]]
                    if self.empt_tol_optn eq 0 then index2 = where((bgkid[index1] eq tmp1) or (abs(bgkid[index1]-kidney[i]) le self.macs_tolerance[1])) $
                    else index2 = where(abs(bgkid[index1]-tmp1) le self.macs_tolerance[1])
                    if self.samp_typ eq 0 then index = index1[index2] $  ;powder sample, A3 is irrelevant
                    else begin
                       tmp1 = min(abs(bgang[index1[index2]]-psi[i]),imin) & tmp1 = bgang[index1[index2[imin]]]
                       if self.empt_tol_optn eq 0 then index3 = where((bgang[index1[index2]] eq tmp1) or (abs(bgang[index1[index2]]-psi[i]) le self.macs_tolerance[2])) $
                       else index3 = where(abs(bgang[index1[index2]]-tmp1) le self.macs_tolerance[2])
                       index = index1[index2[index3]]
                    endelse
                 endif else begin                 ;search order: kidney->A3->En
                    tmp1 = min(abs(bgkid-kidney[i]),imin) & tmp1 = bgkid[imin]
                    if self.empt_tol_optn eq 0 then index1 = where((bgkid eq tmp1) or (abs(bgkid-kidney[i]) le self.macs_tolerance[1])) $
                    else index1 = where(abs(bgkid-tmp1) le self.macs_tolerance[1])
                    if self.samp_typ eq 1 then begin  ;single crystal sample, search A3
                       tmp1 = min(abs(bgang[index1]-psi[i]),imin) & tmp1 = bgang[index1[imin]]
                       if self.empt_tol_optn eq 0 then index2 = where((bgang[index1] eq tmp1) or (abs(bgang[index1]-psi[i]) le self.macs_tolerance[2])) $
                       else index2 = where(abs(bgang[index1]-tmp1) le self.macs_tolerance[2])
                       index1 = index1[index2]
                    endif
                    tmp1 = min(abs(bgen[index1]-en[i]),imin) & tmp1 = bgen[index1[imin]]
                    if self.empt_tol_optn eq 0 then index2 = where((bgen[index1] eq tmp1) or (abs(bgen[index1]-en[i]) le self.macs_tolerance[0])) $
                    else index2 = where(abs(bgen[index1]-tmp1) le self.macs_tolerance[0])
                    index = index1[index2]
                 endelse
                 if n_elements(index) eq 1 then begin
                    qty[*,*,i] = qty[*,*,i]-bgqty[*,*,index[0]]*ssfac
                    err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr[*,*,index[0]])^2*(ssfac*ssfac))
                 endif else begin
                    qty[*,*,i] = qty[*,*,i]-total(bgqty[*,*,index],3)/n_elements(index)*ssfac
                    err[*,*,i] = sqrt(err[*,*,i]^2+total((bgerr[*,*,index])^2,3)/n_elements(index)*(ssfac*ssfac))
                 endelse
                 if keyword_set(debug) then print,'E='+dm_to_string(en[i])+','+dm_to_string(bgen[index])+' kid='+dm_to_string(kidney[i])+','+dm_to_string(bgkid[index])+' A3='+dm_to_string(psi[i])+','+dm_to_string(bgang[index])
             endfor
          endif else begin
             if self.samp_typ ge 2 then begin    ;diffuse scattering type
                if n_bgfile eq 1 then begin
                   if self.samp_typ eq 2 then begin
                      for i=0L,n_file-1L do begin
                          qty[*,i] = qty[*,i]-bgqty[*,0]*ssfac
                          err[*,i] = sqrt(err[*,i]^2+(bgerr[*,0])^2*(ssfac*ssfac))
                      endfor
                   endif else begin
                      for i=0L,n_file-1L do begin
                          qty[*,*,i] = qty[*,*,i]-bgqty[*,*,0]*ssfac
                          err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr[*,*,0])^2*(ssfac*ssfac))
                          if tchan_count ne 0 then begin   ;avoid double-subtraction
                             qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty[tchan_ind,*,0]*ssfac
                             ;error bar is not changed
                          endif
                      endfor
                   endelse
                endif else begin
                   for i=0L,n_file-1L do begin
                       tmp1 = min(abs(bgang-(*self.dataStrPtr).rotation[i]),imin) & tmp1 = bgang[imin]
                       if self.empt_tol_optn eq 0 then index = where((bgang eq tmp1) or (abs(bgang-(*self.dataStrPtr).rotation[i]) le self.psi_tolerance)) $
                       else index = where(abs(bgang-tmp1) le self.psi_tolerance)
                       if self.samp_typ eq 2 then begin
                          if n_elements(index) eq 1 then begin
                             qty[*,i] = qty[*,i]-bgqty[*,index[0]]*ssfac
                             err[*,i] = sqrt(err[*,i]^2+(bgerr[*,index[0]])^2*(ssfac*ssfac))
                          endif else begin
                             qty[*,i] = qty[*,i]-total(bgqty[*,index],2)/n_elements(index)*ssfac
                             err[*,i] = sqrt(err[*,i]^2+total((bgerr[*,index])^2,2)/n_elements(index)*(ssfac*ssfac))
                          endelse
                       endif else begin
                          if n_elements(index) eq 1 then begin
                             qty[*,*,i] = qty[*,*,i]-bgqty[*,*,index[0]]*ssfac
                             err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr[*,*,index[0]])^2*(ssfac*ssfac))
                             if tchan_count ne 0 then begin   ;avoid double-subtraction
                                qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty[tchan_ind,*,index[0]]*ssfac
                                ;error bar is not changed
                             endif
                          endif else begin
                             qty[*,*,i] = qty[*,*,i]-total(bgqty[*,*,index],3)/n_elements(index)*ssfac
                             err[*,*,i] = sqrt(err[*,*,i]^2+total((bgerr[*,*,index])^2,3)/n_elements(index)*(ssfac*ssfac))
                             if tchan_count ne 0 then begin   ;avoid double-subtraction
                                qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+total(bgqty[*,*,index],3)/n_elements(index)*ssfac 
                                ;error bar is not changed
                             endif
                          endelse
                       endelse
                   endfor
                endelse
             endif else begin
                qty = temporary(qty)-(bgqty)*ssfac
                err = sqrt(temporary(err)^2+(bgerr)^2*(ssfac*ssfac))
                if tchan_count ne 0 then begin
                   qty[tchan_ind,*] = qty[tchan_ind,*]+(bgqty[tchan_ind,*])*ssfac
                   ;error bar is not changed
                endif
             endelse
          endelse
          bgqty = 0 & bgerr = 0    ;free up the memory
       endif
       ;subtract empty can file set 2
       if ptr_valid(self.bgdata2) then begin
          if self.instrname eq 'macs' then begin
             for i=0L,nen-1 do begin
                 if self.macs_bgefirst then begin ;search order: En->kidney->A3
                    tmp1 = min(abs(bgen2-en[i]),imin) & tmp1 = bgen2[imin]
                    if self.empt_tol_optn eq 0 then index1 = where((bgen2 eq tmp1) or (abs(bgen2-en[i]) le self.macs_tolerance[0])) $
                    else index1 = where(abs(bgen2-tmp1) le self.macs_tolerance[0])
                    tmp1 = min(abs(bgkid2[index1]-kidney[i]),imin) & tmp1 = bgkid2[index1[imin]]
                    if self.empt_tol_optn eq 0 then index2 = where((bgkid2[index1] eq tmp1) or (abs(bgkid2[index1]-kidney[i]) le self.macs_tolerance[1])) $
                    else index2 = where(abs(bgkid2[index1]-tmp1) le self.macs_tolerance[1])
                    if self.samp_typ eq 0 then index = index1[index2] $  ;powder sample, A3 is irrelevant
                    else begin
                       tmp1 = min(abs(bgang2[index1[index2]]-psi[i]),imin) & tmp1 = bgang2[index1[index2[imin]]]
                       if self.empt_tol_optn eq 0 then index3 = where((bgang2[index1[index2]] eq tmp1) or (abs(bgang2[index1[index2]]-psi[i]) le self.macs_tolerance[2])) $
                       else index3 = where(abs(bgang2[index1[index2]]-tmp1) le self.macs_tolerance[2])
                       index = index1[index2[index3]]
                    endelse
                 endif else begin                 ;search order: kidney->A3->En
                    tmp1 = min(abs(bgkid2-kidney[i]),imin) & tmp1 = bgkid2[imin]
                    if self.empt_tol_optn eq 0 then index1 = where((bgkid2 eq tmp1) or (abs(bgkid2-kidney[i]) le self.macs_tolerance[1])) $
                    else index1 = where(abs(bgkid2-tmp1) le self.macs_tolerance[1])
                    if self.samp_typ eq 1 then begin  ;single crystal sample, search A3
                       tmp1 = min(abs(bgang2[index1]-psi[i]),imin) & tmp1 = bgang2[index1[imin]]
                       if self.empt_tol_optn eq 0 then index2 = where((bgang2[index1] eq tmp1) or (abs(bgang2[index1]-psi[i]) le self.macs_tolerance[2])) $
                       else index2 = where(abs(bgang2[index1]-tmp1) le self.macs_tolerance[2])
                       index1 = index1[index2]
                    endif
                    tmp1 = min(abs(bgen2[index1]-en[i]),imin) & tmp1 = bgen2[index1[imin]]
                    if self.empt_tol_optn eq 0 then index2 = where((bgen2[index1] eq tmp1) or (abs(bgen2[index1]-en[i]) le self.macs_tolerance[0])) $
                    else index2 = where(abs(bgen2[index1]-tmp1) le self.macs_tolerance[0])
                    index = index1[index2]
                 endelse
                 if n_elements(index) eq 1 then begin
                    qty[*,*,i] = qty[*,*,i]-bgqty2[*,*,index[0]]*ssfac
                    err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr2[*,*,index[0]])^2*(ssfac*ssfac))
                 endif else begin
                    qty[*,*,i] = qty[*,*,i]-total(bgqty2[*,*,index],3)/n_elements(index)*ssfac
                    err[*,*,i] = sqrt(err[*,*,i]^2+total((bgerr2[*,*,index])^2,3)/n_elements(index)*(ssfac*ssfac))
                 endelse
             endfor
          endif else begin
             if self.samp_typ ge 2 then begin    ;diffuse scattering type
                if n_bgfile2 eq 1 then begin
                   if self.samp_typ eq 2 then begin
                      for i=0L,n_file-1L do begin
                          qty[*,i] = qty[*,i]-bgqty2[*,0]*ssfac
                          err[*,i] = sqrt(err[*,i]^2+(bgerr2[*,0])^2*(ssfac*ssfac))
                      endfor
                   endif else begin
                      for i=0L,n_file-1L do begin
                          qty[*,*,i] = qty[*,*,i]-bgqty2[*,*,0]*ssfac
                          err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr2[*,*,0])^2*(ssfac*ssfac))
                          if tchan_count ne 0 then begin   ;avoid double-subtraction
                             qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty2[tchan_ind,*,0]*ssfac
                             ;error bar is not changed
                          endif
                      endfor
                   endelse
                endif else begin
                   for i=0L,n_file-1L do begin
                       tmp1 = min(abs(bgang2-(*self.dataStrPtr).rotation[i]),imin) & tmp1 = bgang2[imin]
                       if self.empt_tol_optn eq 0 then index = where((bgang2 eq tmp1) or (abs(bgang2-(*self.dataStrPtr).rotation[i]) le self.psi_tolerance)) $
                       else index = where(abs(bgang2-tmp1) le self.psi_tolerance)
                       if self.samp_typ eq 2 then begin
                          if n_elements(index) eq 1 then begin
                             qty[*,i] = qty[*,i]-bgqty2[*,index[0]]*ssfac
                             err[*,i] = sqrt(err[*,i]^2+(bgerr2[*,index[0]])^2*(ssfac*ssfac))
                          endif else begin
                             qty[*,i] = qty[*,i]-total(bgqty2[*,index],2)/n_elements(index)*ssfac
                             err[*,i] = sqrt(err[*,i]^2+total((bgerr2[*,index])^2,2)/n_elements(index)*(ssfac*ssfac))
                          endelse
                       endif else begin
                          if n_elements(index) eq 1 then begin
                             qty[*,*,i] = qty[*,*,i]-bgqty2[*,*,index[0]]*ssfac
                             err[*,*,i] = sqrt(err[*,*,i]^2+(bgerr2[*,*,index[0]])^2*(ssfac*ssfac))
                             if tchan_count ne 0 then begin   ;avoid double-subtraction
                                qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+bgqty2[tchan_ind,*,index[0]]*ssfac
                                ;error bar is not changed
                             endif
                          endif else begin
                             qty[*,*,i] = qty[*,*,i]-total(bgqty2[*,*,index],3)/n_elements(index)*ssfac
                             err[*,*,i] = sqrt(err[*,*,i]^2+total((bgerr2[*,*,index])^2,3)/n_elements(index)*(ssfac*ssfac))
                             if tchan_count ne 0 then begin   ;avoid double-subtraction
                                qty[tchan_ind,*,i] = qty[tchan_ind,*,i]+total(bgqty2[*,*,index],3)/n_elements(index)*ssfac 
                                ;error bar is not changed
                             endif
                          endelse
                       endelse
                   endfor
                endelse
             endif else begin
                qty = temporary(qty)-(bgqty2)*ssfac
                err = sqrt(temporary(err)^2+(bgerr2)^2*(ssfac*ssfac))
                if tchan_count ne 0 then begin
                   qty[tchan_ind,*] = qty[tchan_ind,*]+(bgqty2[tchan_ind,*])*ssfac
                   ;error bar is not changed
                endif
             endelse
          endelse
          bgqty2 = 0 & bgerr2 = 0    ;free up the memory
       endif
       ;absorption correction
       if ptr_valid(self.absp_info) then begin
          ninfo = n_elements(*(self.absp_info))
          info  = [(*(self.absp_info))[0],(*(self.absp_info))[3:(ninfo-1)]]
          if info[0] ge 0 and info[0] lt n_elements(self.absp_npts) then begin
             npts = (*self.absp_npts[info[0]])[*,(*(self.absp_info))[1]]  
             if self.instrname eq 'macs' then begin
                current1 = systime(/sec)
                for i=0L,nen-1L do begin
                    if info[0] eq 3 then  info[n_elements(info)-1] = (*(self.absp_info))[ninfo-1]+psi[i]-(90.0-self.psi[3]-self.orie_psi) ;flat plate geometry
                    tmp = dm_calc_shielding(info,([Ei[i],Ef[i]])[self.instrgeom],en[i],kidney[i]-76.0+det_tt,interp=(*(self.absp_info))[2],instrgeom=self.instrgeom,is_dcs=(self.instrname eq 'dcs'),npoints=npts,group_leader=self.tlb,$
                              stopped=stopped,title='Absorption Correction:'+dm_to_string(i+1)+'/'+dm_to_string(nen),/noquestion)
                    if i eq 0 then begin
                       time = (systime(/sec)-current1)*nen/60.
                       if time gt 1 then begin
                          ok = dialog_message(['The absorption calculation will take about '+dm_to_string(time,resolution=1)+' mintues. Do you still want to continue?',$
                               'Try reducing the number of sampling points.'],/question,dialog_parent=self.tlb)
                          if ok eq 'No' then stopped = 1b
                       endif
                    endif
                    if (~stopped) then begin
                       qty[0,*,i] = qty[0,*,i]*tmp
                       qty[1,*,i] = qty[1,*,i]*tmp
                       err[0,*,i] = err[0,*,i]*tmp
                       err[1,*,i] = err[1,*,i]*tmp
                    endif else $
                       break
                endfor   
             endif else begin
                if (self.samp_typ lt 2) or (info[0] ne 3) then $
                   tmp = dm_calc_shielding(info,eief,en,det_tt,interp=(*(self.absp_info))[2],instrgeom=self.instrgeom,is_dcs=(self.instrname eq 'dcs'),npoints=npts,group_leader=self.tlb,stopped=stopped)          
                if self.samp_typ ge 2 then begin    ;diffuse scattering type
                   if self.samp_typ eq 2 then begin
                      for i=0L,n_file-1L do begin
                          if info[0] eq 3 then begin ;flat plate in diffuse scattering mode
                             info[n_elements(info)-1] = (*(self.absp_info))[ninfo-1]+psi[i]-self.orie_psi
                             tmp = dm_calc_shielding(info,eief,en,det_tt,interp=(*(self.absp_info))[2],instrgeom=self.instrgeom,is_dcs=(self.instrname eq 'dcs'),npoints=npts,group_leader=self.tlb,$
                                   stopped=stopped,title='Absorption Correction:'+dm_to_string(i+1)+'/'+dm_to_string(n_file),noquestion=(i gt 0),timefactor=n_file)          
                          endif
                          if (~stopped) then begin
                             qty[*,i] = qty[*,i]*tmp
                             err[*,i] = err[*,i]*tmp
                          endif else $
                             break
                      endfor
                   endif else begin
                      for i=0L,n_file-1L do begin
                          if info[0] eq 3 then begin ;flat plate in diffuse scattering mode
                             info[n_elements(info)-1] = (*(self.absp_info))[ninfo-1]+psi[i]-self.orie_psi
                             tmp = dm_calc_shielding(info,eief,en,det_tt,interp=(*(self.absp_info))[2],instrgeom=self.instrgeom,is_dcs=(self.instrname eq 'dcs'),npoints=npts,group_leader=self.tlb,$
                                   stopped=stopped,title='Absorption Correction:'+dm_to_string(i+1)+'/'+dm_to_string(n_file),noquestion=(i gt 0),timefactor=n_file)          
                          endif
                          if (~stopped) then begin
                             qty[*,*,i] = qty[*,*,i]*tmp
                             err[*,*,i] = err[*,*,i]*tmp
                          endif else $
                             break
                      endfor
                   endelse
                endif else begin
                   if (~stopped) then begin
                      qty = temporary(qty)*tmp
                      err = temporary(err)*tmp
                   endif 
                endelse
             endelse
             if stopped then begin  ;change the menu to absorption->None
                self->my_menu_toggle,'absp'+['None','Cylinder','Annulus','Sphere','Flat'],0
                (*(self.absp_info))[0] = -1
                tmpmesg = 'Absorption correction is disabled.'
                if (self.samp_typ ge 2) and (i gt 0) then $ ;some intensity and error bars have been modified
                   tmpmesg = tmpmesg+' The intensities of some angles have been modified.'    
                if n_elements(mesg) eq 0 then $
                   mesg = tmpmesg $
                else $
                   mesg = [mesg,tmpmesg]  
             endif
             tmp = 1    ;free up the memory
             widget_control,/hourglass
          endif
       endif
       ;efficiency adjustment
       if ptr_valid(self.eff) then begin
          if self.instrname eq 'macs' then begin
             for i=0L,nen-1L do begin
                 qty[*,*,i] = qty[*,*,i]*(*self.eff)
                 err[*,*,i] = err[*,*,i]*(*self.eff)
             endfor
          endif else begin
             if self.samp_typ ge 2 then begin    ;diffuse scattering type
                if self.samp_typ eq 2 then begin
                   for i=0L,n_file-1L do begin
                       qty[*,i] = qty[*,i]/(*self.eff)
                       err[*,i] = err[*,i]/(*self.eff)
                   endfor
                endif else begin
                   for i=0L,n_file-1L do begin
                       for j=0L,nen-1L do begin
                           qty[j,*,i] = qty[j,*,i]/(*self.eff)
                           err[j,*,i] = err[j,*,i]/(*self.eff)
                       endfor
                   endfor
                endelse
             endif else begin
                for i=0L,nen-1L do begin
                    qty[i,*] = qty[i,*]/(*self.eff)
                    err[i,*] = err[i,*]/(*self.eff)
                endfor
             endelse
          endelse
       endif
       ;Debye-Waller adjustment
       if (self.intn_typ eq 4) or (self.intn_typ eq 5) then begin   
          ;ask for u^2
          oldu2 = 0. & u2_existed = 0b
          if n_tags((*self.dataStrPtr).info) ne 0 then begin
             u2names = tag_names((*self.dataStrPtr).info)
             index  = where(u2names eq 'DWU2',count)
             if count ne 0 then begin
                oldu2 = (*self.dataStrPtr).info.dwu2
                u2_existed = 1b
             endif
          endif
          if n_elements(dwu2) ne 0 then begin
             notvalid = 0b
             dwu2 = dwu2[0]       
          endif else $
             notvalid = 1b
          while notvalid do begin
             dwu2 = dm_dialog_input('u^2:',/float,default=oldu2,dialog_parent=self.tlb,title='Debye-Waller correction:',info='exp(u^2*Q^2) will be applied to correct for the Debye-Waller factor.')
             widget_control,/hourglass
             if finite(dwu2) then $
                if dwu2 ge 0 then begin
                   if u2_existed then (*self.dataStrPtr).info.dwu2=dwu2
                   notvalid = 0b ;valid u^2 given
                endif
          endwhile
       endif
       ;kfactor adjustment
       if (self.intn_typ eq 1) or (self.intn_typ eq 3) or (self.intn_typ eq 4) or (self.intn_typ eq 5) then begin   ;detailed balance factor
          kb = 8.617342e-2    ;Boltzmann constant in mev.k^-1
          if self.instrname ne 'macs' then begin
             ;ask for a sample temperature
             old = 300. & t_existed = 0b
             if n_tags((*self.dataStrPtr).info) ne 0 then begin
                tnames = tag_names((*self.dataStrPtr).info)
                index  = where(tnames eq 'TEMPERATURE',count)
                if count ne 0 then begin
                   old = (*self.dataStrPtr).info.temperature
                   t_existed = 1b
                endif
             endif
             if n_elements(temperature) ne 0 then begin
                notvalid = 0b
                temp = temperature[0]       
             endif else $
                notvalid = 1b
             while notvalid do begin
                temp = dm_dialog_input('sample temperature:',/float,default=old,dialog_parent=self.tlb,title='Detailed blance calculation')
                widget_control,/hourglass
                if finite(temp) then $
                   if temp gt 0 then begin
                      if t_existed then (*self.dataStrPtr).info.temperature=temp
                      notvalid = 0b ;valid temperature given
                   endif
             endwhile
             ;calculate detailed balance factor
             dbal = exp(-en/(2.0*kb*temp))
          endif else begin
             dbal = exp(-en/(2.0*kb*data_temperature[*,self.macstemp_typ]))
          endelse
       endif
       
       case self.intn_typ of    ;same case for ((self.instrname eq 'nxspe') and (self.lamcorrnxspe eq 0)) 
          0:  if self.instrname eq 'macs' then $      ;S(Q,omega)
              tmp = (*self.kfactor) $                   
              else $
              tmp = (*self.kfactor)^4                 ;S(Q,omega)
          1:  if self.instrname eq 'macs' then $      ;Symmetrized S(Q,omega)
              tmp = (*self.kfactor)*dbal $
              else $
              tmp = (*self.kfactor)^4*dbal            
          2:  tmp = (*self.kfactor)^3                 ;d2sigma/dOmegadEf
          3:  if self.instrname eq 'macs' then $      ;Chi(Q,omega)
              tmp = (*self.kfactor)*(1-dbal^2) $
              else $
              tmp = (*self.kfactor)^4*(1-dbal^2)      
          4:  tmp = (*self.kfactor)^4*(1-dbal^2)*en   ;GDOS
          5:  tmp = (*self.kfactor)^3*(1-dbal^2)*en   ;VDOS   
          6:  tmp = 1.0                               ;no kfactor adjustment
         else:tmp = 1.0                               ;no kfactor adjustment
       endcase
       if ((self.instrname eq 'nxspe') and (self.lamcorrnxspe eq 1)) then begin ;very VERY unlikely case
          case self.intn_typ of
            0:  tmp = (*self.kfactor)^3               ;S(Q,omega)
            1:  tmp = (*self.kfactor)^3*dbal          ;Symmetrized S(Q,omega)
            2:  tmp = (*self.kfactor)^2               ;d2sigma/dOmegadEf
            3:  tmp = (*self.kfactor)^3*(1-dbal^2)    ;Chi(Q,omega)
            4:  tmp = (*self.kfactor)^3*(1-dbal^2)*en ;GDOS
            5:  tmp = (*self.kfactor)^2*(1-dbal^2)*en ;VDOS   
            6:  tmp = 1.0                             ;no kfactor adjustment
           else:tmp = 1.0                             ;no kfactor adjustment
          endcase
       endif
       if ((self.instrname eq 'nxspe') and (self.lamcorrnxspe eq 2)) then begin
          case self.intn_typ of
            0:  tmp = (*self.kfactor)                 ;S(Q,omega)
            1:  tmp = (*self.kfactor)*dbal            ;Symmetrized S(Q,omega)
            2:  tmp = 1.0                             ;d2sigma/dOmegadEf
            3:  tmp = (*self.kfactor)*(1-dbal^2)      ;Chi(Q,omega)
            4:  tmp = (*self.kfactor)*(1-dbal^2)*en   ;GDOS
            5:  tmp = (1-dbal^2)*en                   ;VDOS   
            6:  tmp = 1.0                             ;no kfactor adjustment
           else:tmp = 1.0                             ;no kfactor adjustment
          endcase
       endif
       if ((self.instrname eq 'nxspe') and (self.lamcorrnxspe eq 3)) then begin
          case self.intn_typ of
            0:  tmp = 1.0                             ;S(Q,omega)
            1:  tmp = dbal                            ;Symmetrized S(Q,omega)
            2:  tmp = 1.0/(*self.kfactor)             ;d2sigma/dOmegadEf
            3:  tmp = 1-dbal^2                        ;Chi(Q,omega)
            4:  tmp = (1-dbal^2)*en                   ;GDOS
            5:  tmp = (1-dbal^2)*en/(*self.kfactor)   ;VDOS   
            6:  tmp = 1.0                             ;no kfactor adjustment
           else:tmp = 1.0                             ;no kfactor adjustment
          endcase
       endif      
       if self.eint_yn and (self.samp_typ ne 2) then $
          tmp = temporary(tmp)*ewid
       ;energy dependent detector efficiency adjustment for direct geometry
       if (self.edep_yn eq 1) and (self.instrgeom eq 0) then begin
          he3_temp = 293.12                                                                     ;assume the temperature is 20C = 293 K
          sigma_acoeff = 1.01325*5333/1.38065/he3_temp/sqrt(81.8042/25.3)*self.edep_det[1]*1e-2 ;5333 barn is the he-3 absorption cross section
          sigma_s = 1.01325*6/1.38065/he3_temp*self.edep_det[1]*1e-2                            ;6 barn is the he-3 total scattering cross section
          wl0 = sqrt(81.8042/(*self.dataStrPtr).eief)                                           ;incident wavelength
          wl1 = wl0*(*self.kfactor)                                                             ;scattered wavelength
          if self.instrname eq 'macs' then wl1 = sqrt(81.8042/Ef)
          sumtd = sigma_acoeff*wl1+sigma_s
          ;deteff_inv=1.0/(1.0-exp(-det_thick*sumtd))
          deteff_inv = sumtd/(sumtd-sigma_s)/(1.0-exp(-self.edep_det[0]*sumtd))                 ;from Jeremy's Equation 9
          sumtd = 0 & wl1 = 0
          tmp = temporary(tmp)*temporary(deteff_inv)
       endif
       if self.samp_typ ge 2 then begin
          if (self.samp_typ eq 2) or (n_elements(tmp) eq 1) then begin      
             qty = temporary(qty)*tmp       ;tmp is a scalor here, qty is a [ndet,nfile] arrray
             err = temporary(err)*tmp
          endif else begin                  ;tmp is a [nen] array, qty is a [nen,ndet,nfile] array
             for i=0L,ndet-1L do begin
                 for j=0L,n_file-1L do begin
                     qty[*,i,j] = qty[*,i,j]*tmp
                     err[*,i,j] = err[*,i,j]*tmp
                 endfor
             endfor
          endelse
       endif else begin
          if n_elements(tmp) ne 1 then begin
             if self.instrname eq 'macs' then begin
                for i=0L,ndet-1L do begin      ;tmp is a [nen] array, qty is a [2,ndet,nen] array, only apply the factor to the spec intensity
                    qty[0,i,*] = qty[0,i,*]*tmp
                    err[0,i,*] = err[0,i,*]*tmp
                endfor   
             endif else begin
                for i=0L,ndet-1L do begin      ;tmp is a [nen] array, qty is a [nen,ndet] array
                    qty[*,i] = qty[*,i]*tmp
                    err[*,i] = err[*,i]*tmp
                endfor
             endelse
          endif else begin
             qty = temporary(qty)*tmp
             err = temporary(err)*tmp
          endelse
       endelse
       tmp = 0    ;free up the memory
       ;energy rebin
       if (self.e_bin[self.samp_typ] gt 1) and (self.samp_typ lt 2) then begin
          dm_grid_ebin,self.e_bin[self.samp_typ],qty,err,en,debug=debug,avgsum=(self.bin_avgsum eq 1)
          dm_grid_ebin,self.e_bin[self.samp_typ],ewid
          if (self.samp_typ eq 0) and (self.instrname eq 'dcs') then $
             dm_grid_ebin,self.e_bin[self.samp_typ],time
       endif
       ;calculate dark angle correction factor
       if (self.samp_typ ge 2) and self.dcor_yn and ptr_valid(self.dcor_info) then begin
          tmp_qty = (*self.dcor_info).factor
          tmp_ang = (*self.dcor_info).psi
          tmp_fac = fltarr(n_file)
          for i=0L,n_file-1L do begin
              tmp1 = abs(tmp_ang-(*self.dataStrPtr).rotation[i])  ;find the closest angle index
              tmp1 = min(tmp1,index)    ;index = (0L)>(value_locate(tmp_ang,(*self.dataStrPtr).rotation[i])) ;sort tmp_ang first if using value_locate
              tmp_fac[i] = tmp_qty[index[0]]
          endfor
          if self.samp_typ eq 2 then begin
             for i=0L,n_file-1L do begin
                 qty[*,i] = qty[*,i]*(tmp_fac[i])
                 err[*,i] = err[*,i]*(tmp_fac[i])
             endfor
          endif else begin
             for i=0L,n_file-1L do begin
                 qty[*,*,i] = qty[*,*,i]*(tmp_fac[i])
                 err[*,*,i] = err[*,*,i]*(tmp_fac[i])
             endfor
          endelse
       endif 
    endif
    
    ;mask a3 for macs
    if (self.instrname eq 'macs') and ptr_valid(self.a3mask) then begin
       tmp = bytarr(nen)
       tmpstrs = (*self.a3mask)
       tmp_a3  = 90.0-psi-([0,self.orie_psi])[ptr_valid(self.dataStrPtr)]  ;from psi back to instrument a3
       ;check for > and >= expression
       repeat begin
          tmp1 = stregex(tmpstrs,'>(=*)[ '+string(9b)+']*(-?[0-9.]+)',/subexpr,length=len)
          if len[0] gt 0 then begin
             tmp2 = dm_to_number(strmid(tmpstrs,tmp1[2],len[2]))
             if len[1] gt 0 then ind = where(tmp_a3 ge tmp2,count) $  ;'>='
             else ind = where(tmp_a3 gt tmp2,count)                   ;'>'
             if count gt 0 then tmp[ind] = 1b
             tmpstrs = strmid(tmpstrs,0,tmp1[0])+strmid(tmpstrs,tmp1[0]+len[0],strlen(tmpstrs)-tmp1[0]-len[0])
          endif
       endrep until len[0] le 0 
       ;check for > and >= expression
       repeat begin  
          tmp1 = stregex(tmpstrs,'<(=*)[ '+string(9b)+']*(-?[0-9.]+)',/subexpr,length=len)
          if len[0] gt 0 then begin
             tmp2 = dm_to_number(strmid(tmpstrs,tmp1[2],len[2]))
             if len[1] gt 0 then ind = where(tmp_a3 le tmp2,count) $  ;'<='
             else ind = where(tmp_a3 lt tmp2,count)                   ;'<'
             if count gt 0 then tmp[ind] = 1b
             tmpstrs = strmid(tmpstrs,0,tmp1[0])+strmid(tmpstrs,tmp1[0]+len[0],strlen(tmpstrs)-tmp1[0]-len[0])
          endif
       endrep until len[0] le 0 
       ;check for : expression
       repeat begin  
          tmp1 = stregex(tmpstrs,'(-?[0-9.]+)[ '+string(9b)+']*:[ '+string(9b)+']*(-?[0-9.]+)',/subexpr,length=len)
          if len[0] gt 0 then begin
             tmp2 = dm_to_number(strmid(tmpstrs,tmp1[1],len[1]))
             tmp3 = dm_to_number(strmid(tmpstrs,tmp1[2],len[2]))
             ind = where((tmp_a3 ge tmp2) and (tmp_a3 le tmp3),count) 
             if count gt 0 then tmp[ind] = 1b
             tmpstrs = strmid(tmpstrs,0,tmp1[0])+strmid(tmpstrs,tmp1[0]+len[0],strlen(tmpstrs)-tmp1[0]-len[0])
          endif
       endrep until len[0] le 0 
       ;check for single values
       tmp1 = strsplit(tmpstrs,'[a-zA-Z '+string(9b)+',:<>]',/regex,/extract,count=count)
       for j=0,count-1 do begin
           ind = where(abs(tmp_a3-dm_to_number(tmp1[j])) lt 0.0001,count)
           if count gt 0 then tmp[ind] = 1b
       endfor
       ind = where(tmp eq 0,count,complement=complement,ncomplement=ncomplement) & tmp=0 &tmp_a3=0
       if count ne 0 then begin
          if count ne nen then begin
             Ei = Ei[ind]
             Ef = Ef[ind]
             en = en[ind]
             psi = psi[ind]
             kidney = kidney[ind]
             if ptr_valid(self.dataStrPtr) then begin
                qty = qty[*,*,ind]
                err = err[*,*,ind]
                data_temperature = data_temperature[ind,*]
             endif
             nen = count
          endif
       endif else begin
          ok = dialog_message('All A3 were masked. The mask has been cleared.',dialog_parent=self.tlb,/center)
          self->clear_mask,/a3
       endelse   
    endif
    
    ;mask detectors
    if ptr_valid(self.mask) then begin
       tmp = bytarr(ndet)
       if (self.instrname eq 'macs') and (self.masktype ne 0) then tmp = bytarr(nen*ndet)
       if self.masktype ne 2 then begin
          tmp[*self.mask] = 1b
       endif else if ptr_valid(self.dataStrPtr) or (self.instrname ne 'macs') then begin
          if self.instrname eq 'macs' then begin  ;data available
             tmp_det_tt = det_tt#(fltarr(nen)+1)
             for i=0L,ndet-1 do tmp_det_tt[i,*] = kidney-76.0+det_tt[i]
          endif else $
             tmp_det_tt = det_tt
          tmpstrs = (*self.mask)
          ;check for > and >= expression
          repeat begin
             tmp1 = stregex(tmpstrs,'>(=*)[ '+string(9b)+']*(-?[0-9.]+)',/subexpr,length=len)
             if len[0] gt 0 then begin
                tmp2 = dm_to_number(strmid(tmpstrs,tmp1[2],len[2]))
                if len[1] gt 0 then ind = where(tmp_det_tt ge tmp2,count) $  ;'>='
                else ind = where(tmp_det_tt gt tmp2,count)                   ;'>'
                if count gt 0 then tmp[ind] = 1b
                tmpstrs = strmid(tmpstrs,0,tmp1[0])+strmid(tmpstrs,tmp1[0]+len[0],strlen(tmpstrs)-tmp1[0]-len[0])
             endif
          endrep until len[0] le 0 
          ;check for > and >= expression
          repeat begin  
             tmp1 = stregex(tmpstrs,'<(=*)[ '+string(9b)+']*(-?[0-9.]+)',/subexpr,length=len)
             if len[0] gt 0 then begin
                tmp2 = dm_to_number(strmid(tmpstrs,tmp1[2],len[2]))
                if len[1] gt 0 then ind = where(tmp_det_tt le tmp2,count) $  ;'<='
                else ind = where(tmp_det_tt lt tmp2,count)                   ;'<'
                if count gt 0 then tmp[ind] = 1b
                tmpstrs = strmid(tmpstrs,0,tmp1[0])+strmid(tmpstrs,tmp1[0]+len[0],strlen(tmpstrs)-tmp1[0]-len[0])
             endif
          endrep until len[0] le 0 
          ;check for : expression
          repeat begin  
             tmp1 = stregex(tmpstrs,'(-?[0-9.]+)[ '+string(9b)+']*:[ '+string(9b)+']*(-?[0-9.]+)',/subexpr,length=len)
             if len[0] gt 0 then begin
                tmp2 = dm_to_number(strmid(tmpstrs,tmp1[1],len[1]))
                tmp3 = dm_to_number(strmid(tmpstrs,tmp1[2],len[2]))
                ind = where((tmp_det_tt ge tmp2) and (tmp_det_tt le tmp3),count) 
                if count gt 0 then tmp[ind] = 1b
                tmpstrs = strmid(tmpstrs,0,tmp1[0])+strmid(tmpstrs,tmp1[0]+len[0],strlen(tmpstrs)-tmp1[0]-len[0])
             endif
          endrep until len[0] le 0 
          ;check for single values
          tmp1 = strsplit(tmpstrs,'[a-zA-Z '+string(9b)+',:<>]',/regex,/extract,count=count)
          for j=0,count-1 do begin
              ind = where(abs(tmp_det_tt-dm_to_number(tmp1[j])) lt 0.0001,count)
              if count gt 0 then tmp[ind] = 1b
          endfor
       endif else begin ;'macs' 
          if n_elements(mesg) eq 0 then $ 
             mesg = 'The detector mask has not been applied.' $
          else $
             mesg = [mesg,'The detector mask has not been applied.']
       endelse
       ind = where(tmp eq 0,count,complement=complement,ncomplement=ncomplement) & tmp=0
       if count ne 0 then begin
          if ptr_valid(self.dataStrPtr) then begin
             if self.instrname eq 'macs' then begin
                switch self.masktype of 
                   0: begin
                      qty = qty[*,ind,*]
                      err = err[*,ind,*]
                      break
                      end
                   1: 
                   2: if ncomplement ne 0 then begin
                      qty = reform(qty,2,nen*ndet)
                      err = reform(err,2,nen*ndet)
                      qty[*,complement] = !values.f_nan
                      err[*,complement] = !values.f_nan
                      qty = reform(qty,2,ndet,nen)
                      err = reform(err,2,ndet,nen)
                      complement = 0
                      end
                   else:
                endswitch
             endif else begin
                if self.samp_typ ge 2 then begin
                   if self.samp_typ eq 2 then begin
                      qty = qty[ind,*]
                      err = err[ind,*]
                   endif else begin
                      qty = qty[*,ind,*]
                      err = err[*,ind,*]
                   endelse
                endif else begin
                   qty = qty[*,ind]
                   err = err[*,ind]
                endelse
                if (self.samp_typ eq 0) and self.ftypecheckmask then qtydet = qtydet[ind]
             endelse
          endif
          if (self.masktype eq 0) or (self.instrname ne 'macs') then begin
             det_tt  = det_tt[ind]
             det_psi = det_psi[ind]
             ndet    = count
          endif
          ind     = 0
       endif else begin
          ok = dialog_message('All detectors were masked. The mask has been cleared.',dialog_parent=self.tlb,/center)
          self->clear_mask
       endelse   
    endif

    if self.samp_typ ge 1 then begin
       ;direct -> reciprocal basis vectors
       abc_r = dm_basis_r(self.latt_parm,error=error)
       if error then return  ;error encountered, no further calcuation
       ;orthogonalize and normalize the crystal orientation axes
       if norm(crossp(self.orie_u,self.orie_v)) eq 0 then begin    ;parallel
          ok = dialog_message('Please check cyrstal orientation.',/error,dialog_parent=self.tlb,/center)
          return          ;no further calculation
       endif
       uvw      = dblarr(3,3)
       uvw[*,0] = abc_r#self.orie_u
       uvw[*,1] = abc_r#self.orie_v
       uvw[*,0] = uvw[*,0]/norm(uvw[*,0],/double)
       uvw[*,1] = uvw[*,1]-uvw[*,0]#(transpose(uvw[*,1])#uvw[*,0])
       uvw[*,1] = uvw[*,1]/norm(uvw[*,1],/double)
       for i=0,2 do uvw[i,2]=uvw[(i+1) mod 3,0]*uvw[(i+2) mod 3,1]-uvw[(i+2) mod 3,0]*uvw[(i+1) mod 3,1]
                 ;cross product, not using crossp() for higher precision
       ;rewrite reciprocal lattice parameters in the (u,v,w) frame
       abc_r = transpose(uvw)#(temporary(abc_r))
       if self.samp_typ eq 1 then begin
          if self.instrname eq 'macs' then begin
             if self.macsintn_typ ne 0 then begin   ;'DIFF' or 'SPEC+DIFF' intensity, force Ef=Ei
                Ef = Ei
                en[*] = 0.0
             endif
          endif else begin
             psi = self.orie_psi
          endelse
          type = 1
       endif else type = 2 
    endif

    ptr_free,self.projStrPtr,self.powd_line
  
    ;user function executed before projection
    if (strlen(userf_str) ne 0) and (self.userf_before eq 1) then begin
       if self.userfwarning then begin
          if lmgr(/runtime) then usertmp = 'User Macro' else usertmp = 'User Macro/Function/Procedure'
          if n_elements(mesg) eq 0 then mesg = usertmp+' is specified.' $
          else mesg = [usertmp+' is specified.','',mesg]
          self.userfwarning = 0b
       endif
       ok = execute(userf_str,1,1)
       if ~ok then begin
          ok = dialog_message(!error_state.msg,/error,dialog_parent=self.tlb,/center)
          return  ;error encountered
       endif
       if n_elements(err) ne n_elements(qty) then begin ;in case qty and err are not both defined in user functions
          if n_elements(err) eq 1 then begin
             err = replicate(err,size(qty,/dimension))
          endif else begin
             help,qty,err,output=tmp
             ok = dialog_message(['Incompatible array sizes of intensity and error bar.','',tmp],/error,dialog_parent=self.tlb,/center)
             return
          endelse
       endif
    endif
    
    max_en = max(en,min=min_en)
    if (self.samp_typ eq 2) then begin
       if finite(self.e_range[0]) then min_en = (min_en)<(self.e_range[0])
       if finite(self.e_range[1]) then max_en = (max_en)>(self.e_range[1])
    endif

    case self.samp_typ of
     0:begin  ;powder
       ndat = n_elements(det_tt)
       ;handle the inherent masks in SPE file data
       if ptr_valid(self.dataStrPtr) and self.ftypecheckmask and (self.instrname ne 'macs') then begin
          ind = where(finite(qtydet),count) & qtydet=0
          if count ne ndat then begin
             qty     = qty[*,ind] 
             err     = err[*,ind]
             det_tt  = det_tt[ind]
             det_psi = det_psi[ind]
             ndat    = count
          endif
       endif
       if self.instrname eq 'macs' then begin
          Q0 = dm_proj_spec_samp(Ei,Ef,det_tt,det_psi,type=0,kidney=kidney,debug=debug)   ;Q[ndat,ndet]
          Q1 = dm_proj_spec_samp(Ei,Ei,det_tt,det_psi,type=0,kidney=kidney,debug=debug)   ;Q[ndat,ndet]
          Q  = transpose([[[transpose(Q0)]],[[transpose(Q1)]]])                           ;Q[2,ndat,ndet] 0-spec 1-diff
       endif else begin
          if ~self.negative2th then det_tt = abs(temporary(det_tt))
          ;combine equivalent detectors
          ind     = sort(det_tt)
          if ptr_valid(self.dataStrPtr) then begin
             qty  = qty[*,ind]
             err  = err[*,ind]
          endif
          det_tt  = det_tt[ind]
          det_psi = det_psi[ind]
          ind     = uniq(det_tt)
          ndet    = n_elements(ind)
          if ndet ne ndat then begin
             j = 0L
             for i=0L,ndet-1 do begin
                 if ptr_valid(self.dataStrPtr) then begin
                    qtmp = 0.0
                    etmp = 0.0
                 endif
                 itmp = 0L
                 while(det_tt[j] eq det_tt[ind[i]]) do begin
                    if ptr_valid(self.dataStrPtr) then begin
                       qtmp = temporary(qtmp)+qty[*,j]
                       etmp = temporary(etmp)+err[*,j]*err[*,j]
                       itmp = itmp+1L
                    endif
                    j = j+1L
                    if j eq ndat then break
                 endwhile
                 if self.bin_avgsum eq 1 then itmp=1L
                 if ptr_valid(self.dataStrPtr) then begin
                    qty[*,i] = qtmp/(itmp>1)
                    err[*,i] = sqrt(etmp)/(itmp>1)
                 endif
             endfor
             if ptr_valid(self.dataStrPtr) then begin
                qtmp = 0 & etmp = 0  ;free up the memory
                qty  = qty[*,0:ndet-1]
                err  = err[*,0:ndet-1]
             endif
             det_tt  = det_tt[ind]
             det_psi = det_psi[ind]
          endif
          ind = 0 
          Q = dm_proj_spec_samp(eief,en,det_tt,det_psi,type=0,debug=debug,instrgeom=self.instrgeom)   ;Q[nen,ndet]
       endelse
       if self.intn_typ eq 4 then begin ;GDOS
          factor = 0.48259649/Q^2
          qty = qty*factor*exp(dwu2*Q^2)
          err = err*temporary(factor)
       endif
       if self.intn_typ eq 5 then begin  ;VDOS
          tthmax = max(det_tt,min=tthmin)
          if self.instrgeom eq 0 then $
             efei = eief-en $
          else $
             efei = eief+en
          root = sqrt(eief*efei)
          q4maxmin = (eief+efei-2.0*root*cos(tthmax*!dtor))^2-(eief+efei-2.0*root*cos(tthmin*!dtor))^2
          factor = (1./q4maxmin)#sin(det_tt*!dtor)
          qty = qty*factor*exp(dwu2*Q^2)
          err = err*temporary(factor)
       endif
       if ptr_valid(self.powd_qvalue) and ((min_en le 0) and (max_en ge 0)) then begin  ;powder line
          if self.instrname eq 'macs' then $
             maxQ = max(dm_proj_spec_samp(Ei,Ei,det_tt,det_psi,type=0,kidney=kidney),min=minQ) $
          else $
             maxQ = max(abs(sqrt(eief/2.0721246)*sin(det_tt/2*!dtor)*2.),min=minQ) 
          yval = [-2e27,1e-20,2e27]
          for i=0L,n_elements(*self.powd_qvalue)-1 do begin
             if ((*self.powd_qvalue)[i] lt minQ) then continue
             if ((*self.powd_qvalue)[i] gt maxQ) then break
             if n_elements(powd_x) ne 0 then begin
                powd_x = [powd_x,(*self.powd_qvalue)[i],(*self.powd_qvalue)[i],(*self.powd_qvalue)[i]]
                powd_y = [powd_y,yval]
             endif else begin
                powd_x = [(*self.powd_qvalue)[i],(*self.powd_qvalue)[i],(*self.powd_qvalue)[i]]
                powd_y = yval 
             endelse
             yval[[0,2]] = -yval[[0,2]]
          endfor
          if n_elements(powd_x) ne 0 then self.powd_line=ptr_new([[powd_x],[powd_y]],/no_copy)
       endif
       end
     else:begin  ;single crystal and diffuse scattering
       if self.fold_type eq 1 then begin  ;symmetry rotation folding
          if self.view_u1fold then begin
             index = where(finite(*(self.view_u1foldcenter)),count)
             if count eq 0 then begin
                if n_elements(mesg) eq 0 then $
                   mesg = ['n-fold symmetry value is invalid.'] $
                else $
                   mesg = [mesg,'n-fold symmetry value is invalid.']  
             endif else $
                n_fold = (*(self.view_u1foldcenter))[index[0]]
          endif
       endif
       ;reconstruct the viewing axis and check orthogonality
       view_u1 = self.view_u1
       view_u2 = self.view_u2
       if extravaxis then view_u3 = self.view_u3
       U1    = abc_r#view_u1
       U2    = abc_r#view_u2
       itmp  = where(abs(U1) lt 1e-15,count) & if count gt 0 then U1[itmp] = 0d   ;to avoid the floating underflow condition
       itmp  = where(abs(U2) lt 1e-15,count) & if count gt 0 then U2[itmp] = 0d   ;to avoid the floating underflow condition
       norm1 = norm(U1,/double) & norm2 = norm(U2,/double)
       if self.view_ortharbi eq 0 then begin    ;check orthogonality
          if abs(transpose(U1)#U2/norm1/norm2) gt 1e-8 then begin      ;not orthogonal
             ok = dialog_message('Viewing axes are not orthogonal.',/error,dialog_parent=self.tlb,/center)
             return  ;no further calculation
          endif
       endif else begin
          ;make sure the viewing axes are not parallel to each other
          if norm(crossp(view_u1,view_u2)) eq 0 then begin
             ok = dialog_message('Parallel viewing axes',/error,dialog_parent=self.tlb,/center)
             return  ;no further calculation
          endif
       endelse
       if n_elements(view_u3) eq 3 then begin
          U3    = abc_r#view_u3
          itmp  = where(abs(U3) lt 1e-15,count) & if count gt 0 then U3[itmp] = 0d  ;to avoid the floating underflow condition
          norm3 = norm(U3,/double)
          if self.view_ortharbi eq 0 then begin    ;check orthogonality
             if max(abs([transpose(U2)#U3/norm2/norm3,transpose(U3)#U1/norm3/norm1])) gt 1e-8 then begin   ;not orthogonal
                ok = dialog_message('Viewing axes are not orthogonal.',/error,dialog_parent=self.tlb,/center)
                return   ;no further calculation
             endif
          endif else begin
             ;make sure no viewing axes are parallel to each other
             if (norm(crossp(view_u1,view_u3)) eq 0) or (norm(crossp(view_u2,view_u3)) eq 0) then begin
                ok = dialog_message('Parallel viewing axes',/error,dialog_parent=self.tlb,/center)
                return   ;no further calculation
             endif
          endelse
       endif else $
          U3 = [0d,0d,0d]
       
       if self.instrname eq 'macs' then begin
          Q = dm_proj_spec_samp(Ei,Ef,det_tt,det_psi,psi=psi,type=type,debug=debug,is_spe=self.psi_spe,kidney=kidney)
                                                               ;Q[3,ndat,ndet] in sample frame
       endif else begin
          Q = dm_proj_spec_samp(eief,en,det_tt,det_psi,psi=psi,type=type,debug=debug,is_spe=self.psi_spe,diffetyp=(self.samp_typ eq 3),instrgeom=self.instrgeom)
                                                               ;Q[3,nen,ndet,nfile] in sample frame
       endelse
       QV = dm_proj_view(temporary(Q),U1,U2,U3,n_fold=n_fold,debug=debug)    ;project onto viewing axis

       ;dark angle trajectory
       if ((self.samp_typ ge 2) or ((self.samp_typ eq 1) and (self.instrname eq 'macs'))) and self.datr_yn and finite(self.datr_psi) and (~ptr_valid(self.dataStrPtr)) then begin  
          ;figure out the detector angle range
          if self.instrname eq 'macs' then begin
             tmp_kid = kidney[uniq(kidney,sort(kidney))]
             n_kid = n_elements(tmp_kid) & n_det = n_elements(det_tt)
             tmp_det_tt = (fltarr(n_kid)+1)#det_tt
             tmp_detpsi = (fltarr(n_kid)+1)#det_psi
             for i=0L,n_det-1 do tmp_det_tt[*,i] = tmp_kid-76.0+det_tt[i]
             tmp_psi = psi[uniq(psi,sort(psi))]
             tmp_en = en[uniq(en,sort(en))]
             if self.macsintn_typ ne 0 then tmp_ei = Ei[uniq(Ei,sort(Ei))]     ;DIFF or SPEC+DIFF
             tmp_kid = 0
          endif else begin
             tmp_det_tt = det_tt
             tmp_detpsi = det_psi
             tmp_psi = psi
             tmp_en = en
          endelse
          if self.psi_spe then begin
             pos_ind = where((abs(tmp_detpsi) gt 90) and (abs(tmp_detpsi) lt 270),n_pos,complement=neg_ind,ncomplement=n_neg)       
             if n_pos gt 0 then pos_max = max(tmp_det_tt[pos_ind],min=pos_min)
             if n_neg gt 0 then neg_max = max(-tmp_det_tt[neg_ind],min=neg_min)
          endif else begin
             pos_ind = where(tmp_det_tt gt 0,n_pos,complement=neg_ind,ncomplement=n_neg)
             if n_pos gt 0 then pos_max = max(tmp_det_tt[pos_ind],min=pos_min)
             if n_neg gt 0 then neg_max = max(tmp_det_tt[neg_ind],min=neg_min)
          endelse
          pos_ind = 0 & neg_ind = 0 & tmp_det_tt = 0 & tmp_detpsi = 0
          for i=0,n_elements(tmp_psi)-1 do begin
              da_tt = -(tmp_psi[i]+self.datr_psi)
              while da_tt gt 180 do da_tt = da_tt-360
              while da_tt lt -180 do da_tt = da_tt+360
              if abs((abs(da_tt)-180)) le 2.5 then begin    ;blocking the incident beam
                 if self.instrname eq 'macs' then begin
                    ind = where(psi eq tmp_psi[i],count)
                    if count eq 0 then continue
                    tmp_kid = kidney[ind] & ind = 0
                    tmp_kid = tmp_kid[uniq(tmp_kid,sort(tmp_kid))]
                    n_kid = n_elements(tmp_kid) & n_det = n_elements(det_tt)
                    tmp_det_tt = (fltarr(n_kid)+1)#det_tt
                    tmp_detpsi = (fltarr(n_kid)+1)#det_psi
                    for j=0L,n_det-1 do tmp_det_tt[*,j] = tmp_kid-76.0+det_tt[j]
                    da_tt  = reform(temporary(tmp_det_tt),n_kid*n_det)
                    da_psi = reform(temporary(tmp_detpsi),n_kid*n_det)
                 endif else begin
                    da_tt  = det_tt 
                    da_psi = det_psi
                 endelse
              endif else begin
                 da_tt  = da_tt+(findgen(11)-5.)/2.        ;dark angle range [-2.5,2.5]
                 da_psi = fltarr(11)
                 tmp_ind = where(da_tt gt 180,tmp_count)
                 while tmp_count gt 0 do begin
                    da_tt[tmp_ind] = da_tt[tmp_ind]-360
                    tmp_ind = where(da_tt gt 180,tmp_count)
                 endwhile
                 tmp_ind = where(da_tt lt -180,tmp_count)
                 while tmp_count gt 0 do begin
                    da_tt[tmp_ind] = da_tt[tmp_ind]+360
                    tmp_ind = where(da_tt lt -180,tmp_count)
                 endwhile
                 tmp_ind1 = temporary(tmp_ind) ;destroy tmp_ind
                 if n_pos gt 0 then begin
                    tmp_ind1 = where((da_tt ge pos_min) and (da_tt le pos_max),n_da_tt1)
                    if n_da_tt1 gt 0 then tmp_ind = temporary(tmp_ind1)
                 endif
                 if n_neg gt 0 then begin
                    tmp_ind2 = where((da_tt ge neg_min) and (da_tt le neg_max),n_da_tt2)
                    if n_da_tt2 gt 0 then begin
                       if n_elements(tmp_ind) gt 0 then $
                          tmp_ind = [tmp_ind,temporary(tmp_ind2)] $
                       else $
                          tmp_ind = temporary(tmp_ind2)
                    endif
                 endif
                 if n_elements(tmp_ind) gt 0 then begin
                    da_tt  = da_tt[tmp_ind]
                    da_psi = da_psi[tmp_ind] 
                    if self.psi_spe then begin ;SPE type PSI 
                       ind_psi = where(da_tt gt 0,count_psi)
                       if count_psi gt 0 then da_psi[ind_psi] = 180.0
                       da_tt = abs(da_tt)
                    endif
                 endif else tmp_ind = temporary(da_tt)
              endelse
              if n_elements(da_tt) gt 0 then begin
                 if (self.instrname eq 'macs') and (self.macsintn_typ ne 0) then $
                 daQ0 = dm_proj_spec_samp(tmp_ei,tmp_en,da_tt,da_psi,psi=tmp_psi[i],type=type,is_spe=self.psi_spe,diffetyp=(self.samp_typ eq 3),instrgeom=0) $
                 else $
                 daQ0 = dm_proj_spec_samp(eief,tmp_en,da_tt,da_psi,psi=tmp_psi[i],type=type,is_spe=self.psi_spe,diffetyp=(self.samp_typ eq 3),instrgeom=self.instrgeom)
                 daQ0 = reform(daQ0,3,n_elements(daQ0)/3)  
                 if n_elements(daQ) eq 0 then daQ = daQ0 else daQ=[[daQ],[daQ0]] 
              endif      
          endfor
          tmp_psi = 0 & tmp_en = 0
          if n_elements(daQ) ne 0 then daQV = dm_proj_view(temporary(daQ),U1,U2,U3,n_fold=n_fold)
       endif
           
       ;check folding
       fold = [-1] & fcnt = [0.0] & fdir = [0]
       if self.view_u1fold and (self.fold_type eq 0) then begin
          nfc = n_elements(*(self.view_u1foldcenter))
          nfd = n_elements(*(self.view_u1folddir))
          if nfc ne nfd then begin
             *(self.view_u1folddir) = (dm_dialog_input('center value '+dm_to_string(*(self.view_u1foldcenter))+':',title='Set u1 folding direction',$
                    default =*(self.view_u1folddir)+1,is_droplist=1+intarr(nfc),droplist_content=ptr_new(['-','auto','+']),xsize=100,dialog_parent=self.tlb,/return_number)-1)
            if ~keyword_set(debug) then begin
               value = (['-','auto','+'])[(*self.view_u1folddir)[0]+1]
               for i=1,nfc-1 do value = value+','+(['-','auto','+'])[(*self.view_u1folddir)[i]+1]
               dm_set_droplist,widget_info(self.projBas,find_by_uname='viewU1Folddir'),select=0,set_value=[value,'auto','+']
            endif
          endif
          index = where(finite(*(self.view_u1foldcenter)),count,complement=index1,ncomplement=count1)
          if count gt 0 then begin
             fold = [fold,intarr(count)]
             fcnt = [fcnt,(*(self.view_u1foldcenter))[index]]
             fdir = [fdir,(*(self.view_u1folddir))[index]]
          endif
          if count1 gt 0 then begin
             if count+count1 eq 1 then tmp = 'u1 folding center value is invalid.' $
             else tmp = 'u1 folding center value #'+dm_to_string(index1+1,separator=', ')+([' is',' are'])[count1 gt 1]+' invalid.'
             if n_elements(mesg) eq 0 then mesg = tmp else mesg = [mesg,tmp]
          endif
       endif
       if self.view_u2fold and (self.fold_type eq 0) then begin
          nfc = n_elements(*(self.view_u2foldcenter))
          nfd = n_elements(*(self.view_u2folddir))
          if nfc ne nfd then begin
             *(self.view_u2folddir) = (dm_dialog_input('center value '+dm_to_string(*(self.view_u2foldcenter))+':',title='Set u2 folding direction',$
                    default =*(self.view_u2folddir)+1,is_droplist=1+intarr(nfc),droplist_content=ptr_new(['-','auto','+']),xsize=100,dialog_parent=self.tlb,/return_number)-1)
            if ~keyword_set(debug) then begin
               value = (['-','auto','+'])[(*self.view_u2folddir)[0]+1]
               for i=1,nfc-1 do value = value+','+(['-','auto','+'])[(*self.view_u2folddir)[i]+1]
               dm_set_droplist,widget_info(self.projBas,find_by_uname='viewU2Folddir'),select=0,set_value=[value,'auto','+']
            endif
          endif
          index = where(finite(*(self.view_u2foldcenter)),count,complement=index1,ncomplement=count1)
          if count gt 0 then begin
             fold = [fold,intarr(count)+1]
             fcnt = [fcnt,(*(self.view_u2foldcenter))[index]]
             fdir = [fdir,(*(self.view_u2folddir))[index]]
          endif
          if count1 gt 0 then begin
             if count+count1 eq 1 then tmp = 'u2 folding center value is invalid.' $
             else tmp = 'u2 folding center value #'+dm_to_string(index1+1,separator=', ')+([' is',' are'])[count1 gt 1]+' invalid.'
             if n_elements(mesg) eq 0 then mesg = tmp else mesg = [mesg,tmp]
          endif
       endif
       if n_elements(view_u3) eq 0 then begin
          tmp = size(QV,/dimension)
          case n_elements(tmp) of  ;throw away the third projection
               2:    QV = QV[0:1,*]
               3:    QV = QV[0:1,*,*]
               4:    QV = QV[0:1,*,*,*]
               else: QV = QV[0:1,*]
          endcase
          if n_elements(daQV) ne 0 then daQV = daQV[0:1,*]
       endif 
       if extravaxis then begin
          if self.view_u3fold and (self.fold_type eq 0) then begin
             nfc = n_elements(*(self.view_u3foldcenter))
             nfd = n_elements(*(self.view_u3folddir))
             if nfc ne nfd then begin
                *(self.view_u3folddir) = (dm_dialog_input('center value '+dm_to_string(*(self.view_u3foldcenter))+':',title='Set u3 folding direction',$
                       default =*(self.view_u3folddir)+1,is_droplist=1+intarr(nfc),droplist_content=ptr_new(['-','auto','+']),xsize=100,dialog_parent=self.tlb,/return_number)-1)
                if ~keyword_set(debug) then begin
                   value = (['-','auto','+'])[(*self.view_u3folddir)[0]+1]
                   for i=1,nfc-1 do value = value+','+(['-','auto','+'])[(*self.view_u3folddir)[i]+1]
                   dm_set_droplist,widget_info(self.projBas,find_by_uname='viewU3Folddir'),select=0,set_value=[value,'auto','+']
                endif
             endif
             index = where(finite(*(self.view_u3foldcenter)),count,complement=index1,ncomplement=count1)
             if count gt 0 then begin
                fold = [fold,intarr(count)+2]
                fcnt = [fcnt,(*(self.view_u3foldcenter))[index]]
                fdir = [fdir,(*(self.view_u3folddir))[index]]
             endif
             if count1 gt 0 then begin
                if count+count1 eq 1 then tmp = 'u3 folding center value is invalid.' $
                else tmp = 'u3 folding center value #'+dm_to_string(index1+1,separator=', ')+([' is',' are'])[count1 gt 1]+' invalid.'
                if n_elements(mesg) eq 0 then mesg = tmp else mesg = [mesg,tmp]
             endif
          endif
       endif
       ;folding of viewing axis
       for i=1,n_elements(fold)-1 do begin
           xdat = QV[fold[i],*,*,*]    
           if n_elements(daQV) ne 0 then daxdat = daQV[fold[i],*] 
           case fdir[i] of 
                -1: begin    ;'-' side, unless no data to the '-' side
                    index = where(xdat gt fcnt[i],count,ncomplement=count1)
                    if count1 eq 0 then begin  ;no data to the '-' side
                       count = 0 & index = -1
                    endif
                    if n_elements(daQV) ne 0 then begin
                       daindex = where(daxdat gt fcnt[i],dacount,ncomplement=dacount1)
                       if dacount1 eq 0 then begin  ;no data to the '-' side
                          dacount = 0 & daindex = -1
                       endif
                    endif
                    end
                 0: begin    ;auto
                    ;determine on which side of the folding center that the data has wider range
                    ;fold to greater range side   
                    index = where(xdat gt fcnt[i],count,complement=index1,ncomplement=count1)
                    if (count ne 0) and (count1 ne 0) then begin
                       x_max = max(xdat[index],min=x_min)
                       p_ran = x_max-x_min
                       x_max = max(xdat[index1],min=x_min)
                       n_ran = x_max-x_min  
                       if p_ran ge n_ran then begin
                          index = temporary(index1) 
                          count = count1
                       endif else begin
                          index1 = -1
                       endelse
                    endif else begin
                       tmp = temporary(p_ran) & tmp = temporary(n_ran) & count = 0
                    endelse
                    if n_elements(daQV) ne 0 then begin
                       daindex = where(daxdat gt fcnt[i],dacount,complement=daindex1,ncomplement=dacount1)
                       if (n_elements(p_ran) ne 0) and (n_elements(n_ran) ne 0) then begin
                          if p_ran ge n_ran then begin
                             daindex = temporary(daindex1) 
                             dacount = dacount1
                          endif else begin
                             daindex1 = -1
                          endelse        
                       endif 
                    endif
                    end
                 1: begin    ;'+' side,unless no data to the '+' side
                    index = where(xdat lt fcnt[i],count,ncomplement=count1)
                    if count1 eq 0 then begin  ;no data to the '+' side
                       count = 0 & index = -1
                    endif
                    if n_elements(daQV) ne 0 then begin
                       daindex = where(daxdat lt fcnt[i],dacount,ncomplement=dacount1)
                       if dacount1 eq 0 then begin  ;no data to the '+' side
                          dacount = 0 & daindex = -1
                       endif
                    endif
                    end
              else: count = 0
           endcase
           if count ne 0 then begin
              xdat[index] = (2.*fcnt[i]-xdat[index]) & index = -1
              QV[fold[i],*,*,*] = temporary(xdat)
           endif
           if n_elements(daQV) ne 0 then begin
              if dacount ne 0 then begin
                 daxdat[daindex] = (2.*fcnt[i]-daxdat[daindex]) & daindex = -1
                 daQV[fold[i],*] = temporary(daxdat)
              endif
           endif
       endfor

       ;Al, Cu, or Stainless Steel powder line calculation
       if ptr_valid(self.powd_qvalue) and (min_en le 0) and (max_en ge 0) then begin  ;must be elastic
          tmp_ei = eief
          if self.instrname eq 'macs' then begin
             if self.macsintn_typ eq 0 then begin       ;'SPEC' 
                maxQ = max(dm_proj_spec_samp(Ei,Ef,det_tt,det_psi,type=0,kidney=kidney),min=minQ) 
             endif else begin                           ;'DIFF' or 'SPEC+DIFF'
                maxQ = max(dm_proj_spec_samp(Ei,Ei,det_tt,det_psi,type=0,kidney=kidney),min=minQ)
                tmp_ei = Ei[uniq(Ei,sort(Ei))]
             endelse
          endif else begin
             maxQ = max(abs(sqrt(eief/2.0721246)*sin(det_tt/2*!dtor)*2.),min=minQ)
          endelse
          if self.powd_showall then begin
             tmppsi = findgen(361)*!dtor
          endif else begin
             psimax = max(psi,min=psimin) & npsi = 11>(round(psimax-psimin)+1)
             tmppsi = (psimin-0.5+findgen(npsi)/(npsi-1)*(psimax-psimin+1))*!dtor
          endelse
          tmpcospsi = cos(tmppsi)
          tmpsinpsi = sin(temporary(tmppsi))
          ;now figure out which detectors will intercept powder cones,use 0.5 degree as threshhold
          if self.instrname eq 'macs' then begin
             tmp_kid = kidney[uniq(kidney,sort(kidney))]
             n_kid = n_elements(tmp_kid) & n_det = n_elements(det_tt)
             tmp_det_tt = (fltarr(n_kid)+1)#det_tt
             tmp_detpsi = (fltarr(n_kid)+1)#det_psi
             for i=0L,n_det-1 do tmp_det_tt[*,i] = tmp_kid-76.0+det_tt[i]
             tmp_kid = 0
          endif else begin
             tmp_det_tt = det_tt
             tmp_detpsi = det_psi
          endelse
          for i=0L,n_elements(*self.powd_qvalue)-1 do begin
              if ((*self.powd_qvalue)[i] lt minQ) then continue
              if ((*self.powd_qvalue)[i] gt maxQ) then break
              for j=0L,n_elements(tmp_ei)-1 do begin
                  k = sqrt(tmp_ei[j]/2.07212)
                  tth = asin((*self.powd_qvalue)[i]/2./k)*2/!dtor
                  ind = where(abs(abs(tmp_det_tt)-tth) le 0.5,count)
                  if count eq 0 then continue
                  tmptth = tth*(2*(tmp_det_tt[ind] gt 0)-1)
                  tmpdetpsi = tmp_detpsi[ind]
                  tmp = dm_to_string(tmptth,resolution=3)+'_'+dm_to_string(tmpdetpsi,resolution=3)
                  ind = sort(tmp)
                  ind1 = uniq(tmp[ind])
                  tmptth = tmptth[ind[ind1]]
                  tmpdetpsi = tmpdetpsi[ind[ind1]]
                  for l=0L,n_elements(tmptth)-1 do begin
                      ;in psi=0 uvw frame
                      x = k*(1.0-cos(tmptth[l]*!dtor))
                      sin2th = sin(tmptth[l]*!dtor)
                      sinpsi = sin(tmpdetpsi[l]*!dtor)
                      cospsi = cos(tmpdetpsi[l]*!dtor)
                      if self.psi_spe then begin ;SPE type detector info
                         y = -k*sin2th*cospsi
                         z = -k*sin2th*sinpsi
                      endif else begin           ;DCS type detector info
                         y = k*sin2th*sqrt(1-(sinpsi/sin2th)^2)
                         z = -k*sinpsi
                         if sin2th eq 0 then begin
                            y = 0.0
                            z = 0.0
                         endif
                      endelse
                      ;rotate to sample frame
                      xs = x*tmpcospsi+y*tmpsinpsi
                      ys = y*tmpcospsi-x*tmpsinpsi
                      zs = z+fltarr(n_elements(xs))
                      Vpow = dm_proj_view(transpose([[xs],[ys],[zs]]),U1,U2,U3,n_fold=n_fold)
                      xs = reform(Vpow[0,*]) & ys = reform(Vpow[1,*]) & zs = reform(Vpow[2,*])
                      if n_elements(lines) ne 0 then lines=[[[lines]],[[[xs],[ys],[zs]]]] $
                      else lines=[[xs],[ys],[zs]]
                  endfor      
              endfor
          endfor
          if n_elements(lines) ne 0 then self.powd_line = ptr_new(lines,/no_copy)
       endif
       end
    endcase
    
    ;user function executed after projection
    if (strlen(userf_str) ne 0) and (self.userf_before eq 0) then begin
       if self.userfwarning then begin
          if lmgr(/runtime) then usertmp = 'User Macro' else usertmp = 'User Macro/Function/Procedure'
          if n_elements(mesg) eq 0 then mesg = usertmp+' is specified.' $
          else mesg = [usertmp+' is specified.','',mesg]
          self.userfwarning = 0b
       endif
       ok = execute(userf_str,1,1)
       if ~ok then begin
          ok = dialog_message(!error_state.msg,/error,dialog_parent=self.tlb,/center)
          return  ;error encountered
       endif
       if n_elements(err) ne n_elements(qty) then begin ;in case qty and err are not both defined in user functions
          if n_elements(err) eq 1 then begin
             err = replicate(err,size(qty,/dimension))
          endif else begin
             help,qty,err,output=tmp
             ok = dialog_message(['Incompatible array sizes of intensity and error bar.','',tmp],/error,dialog_parent=self.tlb,/center)
             return
          endelse
       endif
    endif

    if self.instrname eq 'macs' then begin
       if ptr_valid(self.dataStrPtr) then begin
          if size(qty,/n_dimensions) eq 2 then begin
             qty = reform(qty,[size(qty,/dimensions),1])
             err = reform(err,[size(err,/dimensions),1])
          endif
          qty = transpose(qty,[0,2,1])  ;[2,20,ndat]->[2,ndat,ndet]
          err = transpose(err,[0,2,1])
       endif
    endif

    ;store projection
    if self.samp_typ eq 0 then begin        ;powder
       if n_elements(kidney) ne 0 then $
          self.projStrPtr = ptr_new({Q:temporary(Q),en:temporary(en),two_theta:temporary(det_tt),qty:temporary(qty),err:temporary(err),ewid:temporary(ewid),kidney:temporary(kidney),temperature:temporary(data_temperature)}) $ 
       else $
          self.projStrPtr = ptr_new({Q:temporary(Q),en:temporary(en),two_theta:temporary(det_tt),qty:temporary(qty),err:temporary(err),ewid:temporary(ewid),time:temporary(time)}) 
    endif else begin                        ;single crystal    
       if n_elements(daQV) ne 0 then $
          self.projStrPtr = ptr_new({V:temporary(QV),en:temporary(en),ewid:temporary(ewid),two_theta:temporary(det_tt),qty:temporary(qty),err:temporary(err),temperature:temporary(data_temperature),daQV:temporary(daQV)}) $
       else $
          self.projStrPtr = ptr_new({V:temporary(QV),en:temporary(en),ewid:temporary(ewid),two_theta:temporary(det_tt),qty:temporary(qty),err:temporary(err),temperature:temporary(data_temperature)})
    endelse

    ;allow plots
    if (~ keyword_set(debug)) then $
       self->set_display_button  $
    else $
       print,'**Projection calculation finished in ',systime(/sec)-current,' sec.'
    if ptr_valid(self.dataStrPtr) and (self.bin_avgsum eq 1) and (self.e_bin[self.samp_typ] gt 1) and (self.samp_typ lt 2)then begin
       if n_elements(mesg) eq 0 then mesg='The current binning method is SUM.' $
       else mesg = [mesg,'','The current binning method is SUM.']
    endif   
    if n_elements(mesg) ne 0 then ok = dialog_message([mesg,'','This message is only a reminder.'],dialog_parent=self.tlb,/center)
    self.error = 0       ;clear the error flag
end