; $Id: dm_gaussfit.pro,v 1.4.6.1 2013/05/06 17:56:03 ymqiu Exp $
;#######################################################################
;
; NAME:
;  dm_gaussfit
;
; PURPOSE:
;  fit using gaussfit
;
; CATEGORY:
;  dcs_mslice
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  April,2013
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

;this procedure calls gaussfit with reasonable estimates, suitable for fitting just one gaussian
;Parameters:
;   y:           y-values, an [ny] array
;Keywords:
; input:
;   x:           x-values, an [ny] array, if not present, using the index array
;   nterms:      same as the nterms for gaussfit, default is 4 
;   debug:       if set, show debugging info
;   plot:        if set, the fit is shown as a plot
; output:
;   fitnotgood:  returns 1 if the fit is determined as no good
;   chisq:       chisq of gaussfit
;   ny:          n_elements of y
;   params:      [A0,A1,A2,[A3],[A4],[A5]] y=A0*exp[-(x-A1)^2/(2*A2^2)]+A3+A4*x+A5*x^2
  
pro dm_gaussfit,y,x=x,ny=ny,params=params,fitnotgood=fitnotgood,nterms=nterms,chisq=chisq,plot=plot,upsidedown=upsidedown,_ref_extra=extra,debug=debug
    if n_elements(nterms) eq 0 then nterms=4    ;default is 4 terms
    if nterms gt 6 or nterms lt 3 then nterms=4
    ny   = n_elements(y)
    ymax = max(y,imax,min=ymin,subscript_min=imin) 
    if ymax eq ymin then begin
       params = fltarr(nterms)
       if nterms ne 3 then params[3]=ymin
       fitnotgood = 1
       return
    endif
    if n_elements(x) ne ny then begin
       x  = findgen(ny)
       xw = 1.0
    endif else begin
       xw = x[1:ny-1]-x[0:ny-2] & xw=[xw[0],xw]
    endelse    
    if keyword_set(upsidedown) then ximax = x[imin] else ximax = x[imax]
    fwhm  = total((y-ymin)*xw)/(ymax-ymin)
    estimates = fltarr(nterms)
    if nterms eq 3 then begin
       if keyword_set(upsidedown) then estimates[0:2] = [ymin,ximax,fwhm/sqrt(8.*alog(2))] $
       else estimates[0:2] = [ymax,ximax,fwhm/sqrt(8.*alog(2))] 
    endif else begin
       if keyword_set(upsidedown) then estimates[0:3] = [ymin-ymax,ximax,fwhm/sqrt(8.*alog(2)),ymax]$
       else estimates[0:3] = [ymax-ymin,ximax,fwhm/sqrt(8.*alog(2)),ymin]
    endelse
    result = gaussfit(x,y,params,estimates=estimates,nterms=nterms,chisq=chisq,_extra=extra)
    params[2] = abs(params[2])
    if keyword_set(debug) then begin
       print,'estimates=',estimates
       print,'params=',params
       window,/free
       plot,x,y,psym=6,xr=params[1]+[-params[2],params[2]]*3.5,title='chisq='+strtrim(string(chisq),2)
       oplot,x,result
       wait,2 & wdelete,!D.WINDOW
    endif
    ;check the fitting
    if arg_present(fitnotgood) then begin
       if total(finite(params)) ne nterms then fitnotgood = 1b else begin
          minxwid = min(x[1:ny-1]-x[0:ny-2])
          if (abs(params[1]-ximax) gt abs((x[ny-1]-x[0])/4.)) or (params[2] le minxwid/2.) then $
             fitnotgood = 1b $
          else $
             fitnotgood = 0b
       endelse
    endif
    if keyword_set(plot) then begin
       tmp = obj_new('dm_plot',x,y,psym='circle',linestyle='no line')
       tmp->getproperty,xran=xran,yran=yran
       nx = (ny>101)
       xx = xran[0]+findgen(nx)/(nx-1.)*(xran[1]-xran[0])
       ;[A0,A1,A2,[A3],[A4],[A5]] y=A0*exp[-(x-A1)^2/(2*A2^2)]+A3+A4*x+A5*x^2
       yy = params[0]*exp(-(xx-params[1])^2/(2*params[2]^2))
       if nterms ge 4 then yy = yy+params[3]
       if nterms ge 5 then yy = yy+params[4]*xx
       if nterms ge 6 then yy = yy+params[5]*xx^2
       tmp->add_plot,xx,yy,linestyle='solid',color='green'
       tmp->draw
    endif
end