; $Id: dm_load_macs.pro,v 1.5.2.2 2013/05/29 15:09:25 ymqiu Exp $
;#######################################################################
; 
; NAME:
;  dm_load_macs
;
; PURPOSE:
;  load macs file
;  
; CATEGORY:
;  dcs_mslice
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  May, 2013
; 
; HISTORY:
;  02/2013 V1
; 
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

;read macs data file
;input:
;  file:          full file name string(s)
;  group_leader:  parent ID
;output:
;  data:          data array [nrow,ncol]
;  info:          description of columns, lower case output
;  header:        file header string array
;  deteff:        detector efficiency [2,ndet], [0,*] for spec detector, [1,*] for diff detector
;  maspacing:     [monochromator,analyser] spacing
;  latticeparm:   [a,b,c,alpha,beta,gamma]
;  latticeori:    [ux,uy,uz,vx,vy,vz]
;  comment:       comment
;  instrname:     instrument name, lower case output
;  starttime:     start time stamp
;  all_neg:        flag to indicate that all detectors contain -1, no valid data whatsoever
;  error:         error flag
pro dm_load_macs,file,data,info,header=header,deteff=deteff,maspacing=maspacing,latticeparm=latticeparm,latticeori=latticeori,comment=comment,instrname=instrname,group_leader=group_leader,$
    all_neg=all_neg,error=error
  	if arg_present(comment)     then tmp=temporary(comment) 
  	if arg_present(data)        then tmp=temporary(data)   
  	if arg_present(info)        then tmp=temporary(info)   
  	if arg_present(header)      then tmp=temporary(header)  
  	if arg_present(instrname)   then tmp=temporary(instrname)
  	if arg_present(all_neg)     then tmp=temporary(all_neg)
  	if arg_present(maspacing)   then maspacing=fltarr(2)
  	if arg_present(latticeparm) then latticeparm=fltarr(6)
  	if arg_present(latticeori)  then latticeori=fltarr(6)
  	if arg_present(deteff)      then deteff=fltarr(2,20)
  	zerostring = ['out','beo'] ;strings for 0
  	onestring  = ['in','be']   ;strings for 1
  	twostring  = ['hopg']      ;strings for 2
  	tmp=' ' & header=' '
    for i=0,n_elements(file)-1 do begin
        openr,unit,file[i],/get_lun
        while(~stregex(tmp,'#columns',/boolean,/fold_case)) do begin
            readf,unit,tmp
            if (i eq 0) and (arg_present(maspacing) or arg_present(latticeparm) or arg_present(latticeori) or arg_present(deteff) or arg_present(instrname) or arg_present(comment)) then begin
               tmp1 = strlowcase(strsplit(tmp,' '+string(9b),/extract,count=count))
               if arg_present(deteff) then begin
                  if tmp1[0] eq '#detectorefficiencies' then begin
                     for j=1,n_elements(tmp1)-1 do begin
                         ind = stregex(tmp1[j],'^spec([0-9]+)=',length=length,/fold_case,/subexpr)
                         if ind[0] ne -1 then deteff[0,fix(strmid(tmp1[j],ind[1],length[1]))-1] = float(strmid(tmp1[j],ind[0]+length[0]))
                         ind = stregex(tmp1[j],'^diff([0-9]+)=',length=length,/fold_case,/subexpr)
                         if ind[0] ne -1 then deteff[1,fix(strmid(tmp1[j],ind[1],length[1]))-1] = float(strmid(tmp1[j],ind[0]+length[0]))
                     endfor
                  endif
               endif
               if arg_present(maspacing) and (count gt 1) then begin
                  if tmp1[0] eq '#monospacing' then maspacing[0] = float(tmp1[1]) 
                  if tmp1[0] eq '#anaspacing'  then maspacing[1] = float(tmp1[1]) 
               endif
               if arg_present(latticeparm) and (count gt 6) then begin
                  if tmp1[0] eq '#lattice' then latticeparm[*] = float(tmp1[1:6]) 
               endif
               if arg_present(latticeori) and (count gt 6) then begin
                  if tmp1[0] eq '#orient' then latticeori[*] = float(tmp1[1:6]) 
               endif
               if arg_present(instrname) and (count gt 1) then begin
                  if tmp1[0] eq '#instrname' then instrname = tmp1[1]
               endif
               if arg_present(comment) and (count gt 1) then begin
                  marker = '#exptcomment'
                  if tmp1[0] eq marker then comment = strtrim(strmid(tmp,strpos(strlowcase(tmp),marker)+strlen(marker)),2)
               endif
            endif
            if i eq 0 then header = [header,strtrim(tmp,2)]
        endwhile
        if i eq 0 then header = header[1:n_elements(header)-2]
        tmp1 = strlowcase(strsplit(tmp,' '+string(9b),/extract))
        n_info = n_elements(info)
        n_tmp1 = n_elements(tmp1)
        if n_info eq 0 then info=tmp1[1:n_tmp1-1] else if n_info ne n_tmp1-1 then begin
           ok = dialog_message('The files are of different format',/center,/error,dialog_parent=group_leader)
           error = 1b & free_lun,unit
           return
        endif
        ind_spec = where(stregex(info,'spec[0-2][0-9]',/boolean))
        ind_diff = where(stregex(info,'diff[0-2][0-9]',/boolean))
        while not eof(unit) do begin
              readf,unit,tmp
              if stregex(tmp,'^[ '+string(9b)+']*#',/boolean) then continue
              tmp1 = dm_to_number(strsplit(tmp,' '+string(9b),/extract),zerostring=zerostring,onestring=onestring,twostring=twostring,/double)
              if (total(tmp1[[ind_spec,ind_diff]] lt 0) eq 40) then begin ;-1 in all diff and spec counts, ignore them
                 all_neg = 1b
                 continue  
              endif
              if n_elements(ncol) eq 0 then begin
                 ncol = n_elements(tmp1)
                 data = tmp1
              endif else begin
                 if ncol ne n_elements(tmp1) then begin
                    ok = dialog_message(file[i]+' has inconsistent number of columns.',/center,/error,dialog_parent=group_leader)
                    error = 1b & free_lun,unit
                    return
                 endif
                 data=[[data],[tmp1]]
              endelse
        endwhile
        free_lun,unit
	  endfor
	  if n_elements(data) ne 0 then data = transpose(data)
	  all_neg = keyword_set(all_neg) and (n_elements(data) eq 0)
	  error = 0b
end

;this procedure reads macs file for dcs_mslice program
;Parameters:
;   open_path:    file path, string
;   open_file:    file name(s), string or string array
;Keywords:
; input:
;   ftpobj:       ftp object for ftp files
;   parent:       message widget parent,  default is self.tlb
;   title:        message title
; output:
;   error:        error flag 1 or 0
;   qty:          [2,ndet,ndat] (0,*,*)-spec (1,*,*) diff
;   dqty:         [2,ndet,ndat]
;   a3:           [ndat] 
;   cfx:          [ndat] flag for cfx monitors 0-none 1-MgF2 2-Be 3-PG
;   comment:      comment string
;   deteff:       [2,ndet] detector efficiency [0,*] for spec detector, [1,*] for diff detector, note: the efficiency is more like normalizaiton factor Int*deteff, instead of Int/deteff
;   ei:           [ndat]
;   ef:           [ndat]
;   info:         {latticeparm:latticeparm,latticeori:latticeori,temperature:setpoint}  used for checking file compatibility
;   kidney:       [ndat] kidney angle
;   temperature:  [ndat,3]  sample, control, and setpoint temperature
;   weight:       [ndat] monitor
pro dcs_mslice::dm_load_macs,open_path,open_file,a3=a3,cfx=cfx,comment=comment,deteff=deteff,ei=ei,ef=ef,error=error,ftpobj=ftpobj,info=info,$
    kidney=kidney,parent=parent,qty=qty,dqty=dqty,title=title,temperature=temperature,weight=weight
    cd,current=current
    if n_elements(parent) eq 0 then parent = self.tlb
    if n_elements(title)  eq 0 then title  = 'Loading data ...'
    if n_elements(qty)    ne 0 then tmp    = temporary(qty)   ;destroy qty
    n_files = n_elements(open_file)
    cnt_cfx0 = 0L ;number of files with cfx=0
    
    ;catch and clear possible io errors
    catch, myerror
    if myerror ne 0 then begin
       catch,/cancel
       if obj_valid(mesg) then obj_destroy,mesg
       ok = dialog_message(!error_state.msg,/error)
       cd,current
       return
    end
    
    if obj_valid(ftpobj) then pathsep = '/' else pathsep = self.pathsep
    while(strmid(open_path,0,1,/reverse_offset) eq  pathsep) do  open_path = strmid(open_path,0,strlen(open_path)-1)
    for i=0L,n_files-1L do begin
        file = open_path+pathsep+open_file[i]
        if obj_valid(ftpobj) then begin
           ok = ftpobj->GetFileContent(file,localfilename=self.dirs[2]+self.pathsep+'ftptmp'+open_file[i])
           file = self.dirs[2]+self.pathsep+'ftptmp'+open_file[i]
        endif
        if i eq 0 then begin
           mesg =  obj_new('dm_progress',title=title,message='Loading '+file,group_leader=parent) 
           dm_load_macs,file,data,colinfo,header=header,deteff=deteff,maspacing=maspacing,latticeparm=latticeparm,latticeori=latticeori,group_leader=parent,comment=comment,instrname=instrname,error=error
        endif else begin
           mesg->update,message='loading '+file,title='Loading...  '+strtrim(string(i+1),2)+'/'+strtrim(string(n_files),2)
           dm_load_macs,file,data,colinfo,header=header,deteff=deteff,maspacing=maspacing,latticeparm=latticeparm,latticeori=latticeori,group_leader=parent,instrname=instrname,error=error
        endelse 
        if obj_valid(ftpobj) then file_delete,file,/ALLOW_NONEXISTENT,/NOEXPAND_PATH,/QUIET

        if error ne 0 then begin
           if obj_valid(mesg) then obj_destroy,mesg
           cd,current
           return
        endif
        
        if n_elements(data) eq 0 then continue

        if (instrname ne 'ng0') and (instrname ne 'bt9') then exitinfo = open_file[i]+' is not a MACS data file.'
        if n_elements(info) eq 0 then begin
           info = {latticeparm:latticeparm,latticeori:latticeori}
        endif else if self.checklattuv then begin
           if (total(abs(info.latticeparm-latticeparm)) ne 0) or (total(abs(info.latticeori-latticeori)) ne 0) then begin
              exitinfo = 'Selected data files are of different lattice parameters or scattering plane.'
           endif   
        endif
        if n_elements(exitinfo) ne 0 then begin
           ok = dialog_message(exitinfo,/error,/center,dialog_parent=parent)
           cd,current
           if obj_valid(mesg) then obj_destroy,mesg
           error = 1b
           return
        endif
        ;check index every time, since the file format could be different
        ind_qty = lonarr(2,20)
        for j=1,20 do begin
            tmp = string(j,format='(i02)')
            ind_qty[0,j-1] = (where(colinfo eq 'spec'+tmp))[0]
            ind_qty[1,j-1] = (where(colinfo eq 'diff'+tmp))[0]
        endfor
        ind_a3  = (where(colinfo eq 'a3'))[0]
        ind_a2  = (where(colinfo eq 'a2'))[0]
        ind_a6  = (where(colinfo eq 'a6'))[0]
        ind_kn  = (where(colinfo eq 'kidney'))[0]
        ind_wt  = (where(colinfo eq 'monitor'))[0]
        ind_t   = (where(colinfo eq 'temp',cnt_t))[0]
        ind_tc  = (where(colinfo eq 'temperaturecontrolreading',cnt_tc))[0]
        ind_ts  = (where(colinfo eq 'temperaturesetpoint',cnt_ts))[0]
        nstr    = [['mg','be','pg']+'filmon']
        cnt_cfx = 0
        for j=0,n_elements(nstr)-1 do begin
            ind_cfx = (where(colinfo eq nstr[j],count))[0]
            if count ne 0 then begin
               if finite(data[0,ind_cfx]) then begin
                  if round(data[0,ind_cfx]) then begin
                     cnt_cfx = j+1
                     break
                  endif
               endif
            endif
        endfor
        ;check if the date is after CFXdate
        tmp = stregex(header,'#Date(.*)',/fold_case,/subexpr,length=len)
        ind = where(len[1,*] gt 0,count)
        if count gt 0 then begin
           date = dm_to_number(strmid(header[ind[0]],tmp[1,ind[0]],len[1,ind[0]]),/date)
           if (cnt_cfx eq 0) and (date ge self.macs_CFXdate) then cnt_cfx0 = cnt_cfx0+1
        endif
        ;temperature
        if cnt_t ne 0 then begin
           thistemperature = data[*,[ind_t,ind_tc,ind_ts]]
           if self.macstempavg_yn then begin
              thistemperature = (fltarr(n_elements(data[*,0]))+1.)#total(thistemperature,1)/n_elements(thistemperature[*,0])  ;average temperature of the file
           endif
        endif else begin
           if self.extravaxis_yn then begin
              repeat begin
                  ans = dm_dialog_input('T (K)',/float,default=ans,dialog_parent=self.tlb,/align_center,info=open_file[i]+" doesn't contain temperature. Please enter it.",title='Missing temperature')
              endrep until finite(ans)
              thistemperature = ans+fltarr(n_elements(data[*,0]),3)
           endif else $
              thistemperature = -1.0+fltarr(n_elements(data[*,0]),3)
        endelse      
        
        if cnt_ts ne 0 then begin
           tnames = tag_names(info)
           index  = where(tnames eq 'TEMPERATURE',count)
           if count eq 0 then info = create_struct('TEMPERATURE',data[0,ind_ts],info)   
        endif
        if n_elements(qty) eq 0 then begin  
           qty     = transpose([[[data[*,ind_qty[0,*]]]],[[data[*,ind_qty[1,*]]]]])
           a3      = data[*,ind_a3]
           ei      = 81.8042/((2*maspacing[0]*sin(!dtor*data[*,ind_a2]/2))^2)  ;A2->Ei
           ef      = 81.8042/((2*maspacing[1]*sin(!dtor*data[*,ind_a6]/2))^2)  ;A6->Ef
           kidney  = data[*,ind_kn]
           weight  = data[*,ind_wt]  
           temperature = thistemperature         
           if cnt_cfx gt 0 then cfx = cnt_cfx*fix(round(data[*,ind_cfx])) else cfx = intarr(n_elements(data[*,0]))
        endif else begin
           tmp     = transpose([[[data[*,ind_qty[0,*]]]],[[data[*,ind_qty[1,*]]]]])
           qty     = [[[qty]],[[tmp]]]
           a3      = [a3,data[*,ind_a3]]
           ei      = [ei,81.8042/((2*maspacing[0]*sin(!dtor*data[*,ind_a2]/2))^2)]
           ef      = [ef,81.8042/((2*maspacing[1]*sin(!dtor*data[*,ind_a6]/2))^2)]
           kidney  = [kidney,data[*,ind_kn]]
           weight  = [weight,data[*,ind_wt]]    
           temperature = transpose([[transpose(temperature)],[transpose(thistemperature)]])         
           if cnt_cfx gt 0 then cfx = [cfx,cnt_cfx*fix(round(data[*,ind_cfx]))] else cfx = [cfx,intarr(n_elements(data[*,0]))]
        endelse
        if (~ obj_valid(mesg)) then break
    endfor
    
    ;check for NAN's in Ei and Ef, if found, replace them with the corresponding values in Ef and Ei
    if n_elements(data) gt 0 then begin
       ind = where(finite(ei,/nan),count)
       if count gt 0 then ei[ind] = ef[ind]
       ind = where(finite(ef,/nan),count)
       if count gt 0 then ef[ind] = ei[ind]
       dqty = sqrt(qty)
    endif

    cd,current
    if obj_valid(mesg) then obj_destroy,mesg

    ;monitor lambda/2 correction warning
    if cnt_cfx0 ne 0 and ~self.macs_lamb2 then begin
       if n_files eq 1 then mesg = 'No CFX filter is used in the file.' $
       else begin
          if cnt_cfx0 eq n_files then mesg = 'No CFX filter is used in all the files.' $
          else mesg = 'No CFX filter is used in some of the files.'
       endelse
       ok = dialog_message(dialog_parent=self.tlb,mesg+' Do you want to enable the monitor lambda/2 correction?',/question,/center,title='Monitor lambda/2 correction')   
       if ok eq 'Yes' then begin
          self.macs_lamb2 = 1b
          dm_set_button,widget_info(self.tlb,find_by_uname='monlamb2'),self.macs_lamb2,onstring='Apply Monitor Lambda/2 Correction'
       endif
    endif
end