; $Id: dm_volumegrid_bin.pro,v 1.10 2013/03/15 21:57:46 ymqiu Exp $
;#######################################################################
;
; NAME:
;  dm_volumegrid_bin
;
; PURPOSE:
;  bin volume data according to given xstep, ystep, and zstep
;
; CATEGORY:
;  general
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  March, 2013
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

; parameters:
;   in_x(y,z)step:   x(y,z)step
;   xdat:       input xdat[ndat]    output xdat[xsize]
;   ydat:       input ydat[ndat]    output ydat[ysize]
;   zdat:       input zdat[ndat]    output zdat[zsize]
;   idat:       input idat[ndat]    output idat[xsize,ysize,zsize]
; keyword:
;   avgsum:     0: average 1:sum
;   bintr:      binning threshold, bins with fewer data points than bintr*(max number of points in a bin) will be discarded
;   conststep:  0: use the average as the resulted binning values(default) 1: constant binning intervals
;   xstart:   binning starting point, if absent, use the min(xdat) as the starting point
;   ystart:   binning starting point, if absent, use the min(ydat) as the starting point
;   zstart:   binning starting point, if absent, use the min(zdat) as the starting point   
;   checkfinite: flag for checking finite values of zdat
;   leave_blank: if set, leave empty column or rows as they are
;   ewid:       energy width, used when e-integrated intensity option is chose
;               input  dat=sum(I*ewid)
;               output dat=sum(I*ewid)/sum(ewid)
;   extlib:     0:not using external library 1:use external library            
;   ierr:       input error bar of idat[ndat], ouput ierr[xsize,ysize,zsize]
pro dm_volumegrid_bin,in_xstep,in_ystep,in_zstep,xdat,ydat,zdat,idat,ierr=ierr,xstart=xstart,ystart=ystart,zstart=zstart,avgsum=avgsum,bintr=bintr,checkfinite=checkfinite,$
    conststep=conststep,ewid=ewid,leave_blank=leave_blank,extlib=extlib,debug=debug
    if n_elements(avgsum) eq 0 then avgsum=0          ;default using average in binning
    if n_elements(bintr)  eq 0 then bintr=0
    if n_elements(conststep) eq 0 then conststep = 0  ;default is using average as the binned value
    if keyword_set(conststep) then leave_blank=1
    if n_elements(extlib) eq 0 then extlib = 1b       ;default using external binning library if available
    if keyword_set(debug) then current=systime(/sec)
    ndat    = n_elements(idat)
    ok_ierr = n_elements(ierr) eq ndat
    ok_eint = n_elements(ewid) eq ndat
    if keyword_set(checkfinite) then begin
       ind = where(finite(idat),count) 
       if (count ne ndat) and (count ne 0) then begin
          xdat = xdat[ind]
          ydat = ydat[ind]
          zdat = zdat[ind]
          idat = idat[ind]
          if ok_ierr then ierr = ierr[ind]
          if ok_eint then ewid = ewid[ind]
          ind  = 0 & ndat = count
       endif
    endif
    xstep   = abs(in_xstep)
    ystep   = abs(in_ystep)
    zstep   = abs(in_zstep)
    xmax    = max(xdat,min=xmin)
    ymax    = max(ydat,min=ymin)
    zmax    = max(zdat,min=zmin)
    if n_elements(xstart) eq 1 then begin
       if finite(xstart) then xmin = xstart
    endif
    if n_elements(ystart) eq 1 then begin
       if finite(ystart) then ymin = ystart
    endif
    if n_elements(zstart) eq 1 then begin
       if finite(zstart) then zmin = zstart
    endif
    if (xstep eq 0) then xstep = ((xmax eq xmin)?0.1:(xmax-xmin)/100.)
    if (ystep eq 0) then ystep = ((ymax eq ymin)?0.1:(ymax-ymin)/100.)
    if (zstep eq 0) then zstep = ((zmax eq zmin)?0.1:(zmax-zmin)/100.)
    xmax    = xmax+xstep/2.
    xmin    = xmin-xstep/2.
    ymax    = ymax+ystep/2.
    ymin    = ymin-ystep/2.
    zmax    = zmax+zstep/2.
    zmin    = zmin-zstep/2.
    xsize   = (xmax-xmin)/xstep
    ysize   = (ymax-ymin)/ystep
    zsize   = (zmax-zmin)/zstep
    if xsize gt fix(xsize,type=3) then xsize=(2.>(xsize+0.5))
    if ysize gt fix(ysize,type=3) then ysize=(2.>(ysize+0.5))
    if zsize gt fix(zsize,type=3) then zsize=(2.>(zsize+0.5))
    xsize   = fix(xsize,type=3)+1 & ysize = fix(ysize,type=3)+1 & zsize = fix(zsize,type=3)+1
    xlookup = xmin+findgen(xsize+1)*xstep
    ylookup = ymin+findgen(ysize+1)*ystep
    zlookup = zmin+findgen(zsize+1)*zstep
    tmp_x   = fltarr(xsize) & num_x = lonarr(xsize)
    tmp_y   = fltarr(ysize) & num_y = lonarr(ysize)
    tmp_z   = fltarr(zsize) & num_z = lonarr(zsize)
    tmp_i   = fltarr(xsize,ysize,zsize) & num_i = lonarr(xsize,ysize,zsize)
    if ok_ierr then tmp_e = fltarr(xsize,ysize,zsize)
    if ok_eint then tmp_w = fltarr(xsize,ysize,zsize)
    
    defsysv,'!DAVE_AUXILIARY_DIR',exists=exists
    if exists and extlib then begin
       filename = 'rebinnd'+'_'+!version.os+'_'+strtrim(string(!VERSION.MEMORY_BITS),2)
       shlib = (FILE_SEARCH(!DAVE_AUXILIARY_DIR+'rebin_nd'+path_sep()+filename+'*'))[0]
    endif else shlib = ''
    
    if (strlen(shlib) gt 0)then begin
      if (ok_eint and ok_ierr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin3d_iew', long(ndat),float(xdat),float(ydat),float(zdat),$
        double(idat),double(ierr),double(ewid),long(xsize),long(ysize),long(zsize),float(xmin),float(xstep),$
        float(ymin),float(ystep), float(zmin),float(zstep), num_x, tmp_x, num_y, tmp_y, num_z, tmp_z, $
        num_i,tmp_i,tmp_e, tmp_w, /CDECL) 
      end       
      if (ok_eint and ~ok_ierr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin3d_iw', long(ndat),float(xdat),float(ydat),float(zdat),$
        double(idat),double(ewid),long(xsize),long(ysize),long(zsize),float(xmin),float(xstep),$
        float(ymin),float(ystep), float(zmin),float(zstep), num_x, tmp_x, num_y, tmp_y, num_z, tmp_z, $
        num_i,tmp_i, tmp_w, /CDECL)  
      end       
      if (~ok_eint and ok_ierr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin3d_ie', long(ndat),float(xdat),float(ydat),float(zdat),$
        double(idat),double(ierr),long(xsize),long(ysize),long(zsize),float(xmin),float(xstep),$
        float(ymin),float(ystep), float(zmin),float(zstep), num_x, tmp_x, num_y, tmp_y, num_z, tmp_z, $
        num_i,tmp_i,tmp_e,  /CDECL) 
      end       
      if (~ok_eint and ~ok_ierr) then begin
        r = CALL_EXTERNAL(shlib, 'rebin3d_i', long(ndat),float(xdat),float(ydat),float(zdat),$
        double(idat),long(xsize),long(ysize),long(zsize),float(xmin),float(xstep),$
        float(ymin),float(ystep), float(zmin),float(zstep), num_x, tmp_x, num_y, tmp_y, num_z, tmp_z, $
        num_i,tmp_i,  /CDECL)
      end 
    end else begin
      for i=0L,ndat-1L do begin
          ix = value_locate(xlookup,xdat[i])
          iy = value_locate(ylookup,ydat[i])
          iz = value_locate(zlookup,zdat[i])
          tmp_x[ix] = tmp_x[ix]+xdat[i] & num_x[ix] = num_x[ix]+1L
          tmp_y[iy] = tmp_y[iy]+ydat[i] & num_y[iy] = num_y[iy]+1L
          tmp_z[iz] = tmp_z[iz]+zdat[i] & num_z[iz] = num_z[iz]+1L
          tmp_i[ix,iy,iz] = tmp_i[ix,iy,iz]+idat[i] & num_i[ix,iy,iz] = num_i[ix,iy,iz]+1L
          if ok_ierr then tmp_e[ix,iy,iz] = tmp_e[ix,iy,iz]+ierr[i]*ierr[i]
          if ok_eint then tmp_w[ix,iy,iz] = tmp_w[ix,iy,iz]+ewid[i]
        endfor
    endelse
    
    indx = where(num_x ne 0,countx,complement=indx1,ncomplement=countx1)
    indy = where(num_y ne 0,county,complement=indy1,ncomplement=county1)
    indz = where(num_z ne 0,countz,complement=indz1,ncomplement=countz1)
    if keyword_set(leave_blank) then begin
       if countx1 ne 0 then begin
          num_x[indx1]=1.0 & tmp_x[indx1]=xlookup[indx1]
       endif
       if county1 ne 0 then begin
          num_y[indy1]=1.0 & tmp_y[indy1]=ylookup[indy1]
       endif
       if countz1 ne 0 then begin
          num_z[indz1]=1.0 & tmp_z[indz1]=zlookup[indz1]
       endif
    endif else begin
       if (countx ne 0) and (countx1 ne 0) then begin
          tmp_x = tmp_x[indx]   & num_x = num_x[indx]
          tmp_i = tmp_i[indx,*,*] & num_i = num_i[indx,*,*]
          if ok_ierr then tmp_e = tmp_e[indx,*,*]
          if ok_eint then tmp_w = tmp_w[indx,*,*]
       endif
       if (county ne 0) and (county1 ne 0) then begin
          tmp_y = tmp_y[indy]   & num_y=num_y[indy]
          tmp_i = tmp_i[*,indy,*] & num_i=num_i[*,indy,*]
          if ok_ierr then tmp_e = tmp_e[*,indy,*]
          if ok_eint then tmp_w = tmp_w[*,indy,*]
       endif
       if (countz ne 0) and (countz1 ne 0) then begin
          tmp_z = tmp_z[indz]   & num_z=num_z[indz]
          tmp_i = tmp_i[*,*,indz] & num_i=num_i[*,*,indz]
          if ok_ierr then tmp_e = tmp_e[*,*,indz]
          if ok_eint then tmp_w = tmp_w[*,*,indz]
       endif
    endelse
    indx=-1 & indx1=-1 & indy=-1 & indy1=-1 & indz=-1 & indz1=-1
    if keyword_set(conststep) then begin
       xdat = xmin+(findgen(xsize-1)+0.5)*xstep
       ydat = ymin+(findgen(ysize-1)+0.5)*ystep
       zdat = zmin+(findgen(zsize-1)+0.5)*zstep
       tmp_i = tmp_i[0:xsize-2,0:ysize-2,0:zsize-2]
       num_i = num_i[0:xsize-2,0:ysize-2,0:zsize-2]
       if ok_ierr then tmp_e = tmp_e[0:xsize-2,0:ysize-2,0:zsize-2]
       if ok_eint then tmp_w = tmp_w[0:xsize-2,0:ysize-2,0:zsize-2]
    endif else begin
       xdat = temporary(tmp_x)/temporary(num_x)
       ydat = temporary(tmp_y)/temporary(num_y)
       zdat = temporary(tmp_z)/temporary(num_z)
    endelse
    idat = temporary(tmp_i)
    if ok_ierr then ierr = sqrt(temporary(tmp_e))
    index1 = where(num_i ne 0,count1,complement=index0,ncomplement=count0)
    if (count1 ne 0) then begin
       if ok_eint then begin
          idat[index1] = idat[index1]/tmp_w[index1]
          if ok_ierr then ierr[index1] = ierr[index1]/tmp_w[index1]
       endif else if avgsum eq 0 then begin   
          idat[index1] = idat[index1]/num_i[index1]    ;average
          if ok_ierr then ierr[index1] = ierr[index1]/num_i[index1]
       endif 
       if (bintr gt 0.0) and (bintr lt 1.0) then begin
          min_num_i = bintr*(max(num_i)>1)
          ind = where(num_i lt min_num_i,ct1)
          if ct1 gt 0 then begin
             idat[ind] = !values.f_nan
             if ok_ierr then ierr[ind] = !values.f_nan
          endif
       endif
    endif
    if (count0 ne 0) then begin
       idat[index0] = !values.f_nan
       if ok_ierr then ierr[index0] = !values.f_nan
    endif
    if keyword_set(debug) then print,'stepgrid_bin finished in ',systime(/sec)-current,' secs.'
end